
import numpy as n
cimport numpy as n
cimport cython

from pyknot2.utils import vprint

from pyknot2.writhes import validate_diagram

@cython.wraparound(False)
@cython.boundscheck(False)
def writhing_numbers_cython(gc, diagrams, based=False):

    if not isinstance(diagrams, (list, tuple)):
        diagrams = [diagrams]

    for d in diagrams:
        validate_diagram(d)

    level = 0

    code = gc._gauss_code
    code = code[0]
    gc_len = len(gc)
    code_len = len(code)
    from pyknot2.invariants import _crossing_arrows_and_signs
    arrows, signs = _crossing_arrows_and_signs(code, gc.crossing_numbers)

    crossing_numbers = list(gc.crossing_numbers)

    # degrees = [len(diagram.split(',')) for diagram in diagrams]

    degrees = defaultdict(lambda: [])
    for diagram in diagrams:
        degrees[len(diagram.split(',')) // 2].append(diagram)

    relations = {diagram: [] for diagram in diagrams}
    for diagram in diagrams:
        degree = len(diagram.split(',')) // 2
        num_relations = factorial(degree - 1) * 4

        terms = diagram.split(',')
        numbers = [term[:-1] for term in terms]

        number_strs = list(sorted(set(numbers), key=lambda j: int(j)))
        for i, number in enumerate(number_strs):
            for oi, other_number in enumerate(number_strs[i+1:]):
                oi += i + 1
                if i != 0:
                    if terms.index(number + '-') < terms.index(other_number + '-'):
                        relations[diagram].append(lambda l, i=i, oi=oi: l[i][0] < l[oi][0])
                    else:
                        relations[diagram].append(lambda l, i=i, oi=oi: l[i][0] > l[oi][0])

                    if terms.index(number + '-') < terms.index(other_number + '+'):
                        relations[diagram].append(lambda l, i=i, oi=oi: l[i][0] < l[oi][1])
                    else:
                        relations[diagram].append(lambda l, i=i, oi=oi: l[i][0] > l[oi][1])

                if terms.index(number + '+') < terms.index(other_number + '-'):
                    relations[diagram].append(lambda l, i=i, oi=oi: l[i][1] < l[oi][0])
                else:
                    relations[diagram].append(lambda l, i=i, oi=oi: l[i][1] > l[oi][0])

                if i == 0:
                    if terms.index(number + '+') < terms.index(other_number + '+'):
                        relations[diagram].append(lambda l, i=i, oi=oi: l[i][1] < l[oi][1])
                    else:
                        relations[diagram].append(lambda l, i=i, oi=oi: l[i][1] > l[oi][1])


    max_degree = max(degrees.keys())

    used_sets = set()

    # representations_sums = [0 for _ in diagrams]
    representations_sums = {d: 0 for d in diagrams}
    used_sets = {d: set() for d in diagrams}

    print('arrows are', arrows)

    try:
        num_combs = (factorial(len(crossing_numbers)) //
                     factorial(max_degree) //
                     factorial(len(crossing_numbers) - max_degree))
    except ValueError:
        num_combs = 0

    combs = combinations(crossing_numbers, max_degree)

    counter = n.zeros(max_degree)
    arrows = 
    while counter[0] < max_degree:
        
    for ci, comb in enumerate(combs):
        if ci % 10000 == 0:
            vprint('\rCombination {} of {}    '.format(ci + 1, num_combs),
                   newline=False)

        if based:
            perms = [comb]
        else:
            perms = permutations(comb)

        ordered_indices = tuple(sorted(comb))
        for diagram in diagrams:
            if ordered_indices not in used_sets[diagram]:
                break
        else:
            continue

        for perm in perms:
            cur_arrows = [list(arrows[i]) for i in perm]

            a1s = cur_arrows[0][0]
            if based:
                a1s = 0

            for i, arrow in enumerate(cur_arrows):
                arrow[0] = (arrow[0] - a1s) % code_len
                arrow[1] = (arrow[1] - a1s) % code_len


            for diagram in diagrams:
                if ordered_indices in used_sets[diagram]:
                    continue
                for relation in relations[diagram]:
                    if not relation(cur_arrows):
                        break
                else:
                    representations_sums[diagram] += (
                        reduce(lambda x, y: x*y,
                               [signs[arrow_i] for arrow_i in perm]))
                    used_sets[diagram].add(ordered_indices)
                        
    vprint()

    return representations_sums



'''
SpaceCurve
----------

The :class:`SpaceCurve` class is the base for all space-curve analysis
in pyknotid. It provides methods for geometrical manipulation
(translation, rotation etc.), calculating geometrical characteristics
such as the writhe, and obtaining topological representations of the
curve by analysing its crossings in projection.

pyknotid provides other classes for topological analysis of the curves:

- :class:`~pyknotid.spacecurves.knot.Knot` for calculating knot invariants of the space-curve.
- :class:`~pyknotid.spacecurves.link.Link` to handle multiple SpaceCurves and calculate linking invariants.
- :class:`~pyknotid.spacecurves.periodiccell.Cell` for handling multiple space-curves in a box with periodic boundaries.

API documentation
~~~~~~~~~~~~~~~~~

'''

import numpy as n
import numpy as np
import sys

try:
    from pyknotid.spacecurves import chelpers
except ImportError:
    print('Could not import cythonised chelpers, using Python '
          'alternative. This will '
          'give the same result, but is slower.')
    from pyknotid.spacecurves import helpers as chelpers
from pyknotid.spacecurves import helpers as helpers
from pyknotid.spacecurves.geometry import arclength, radius_of_gyration
from pyknotid.spacecurves.smooth import smooth

# We must be careful only to import modules that do not depend on this one, to
# prevent import loops
from pyknotid.visualise import plot_line, plot_projection
from pyknotid.io import to_json_file, from_json_file, from_csv
from pyknotid.utils import (mag, get_rotation_matrix,
                           ensure_shape_tuple)


class SpaceCurve(object):
    '''
    Class for holding the vertices of a single line, providing helper
    methods for convenient manipulation and analysis.

    The methods of this class are largely geometrical (though this
    includes listing the crossings in projection and extracting a
    Gauss code etc.). For topological measurements, you should use
    a :class:`~pyknotid.spacecurves.knot.Knot`.

    This class deliberately combines methods to do many different kinds
    of measurements or manipulations. Some of these are externally
    available through other modules in pyknotid - if so, this is usually
    indicated in the method docstrings.

    Parameters
    ----------
    points : array-like
        The 3d points (vertices) of a piecewise
        linear curve representation
    verbose : bool
        Indicates whether the SpaceCurve should print
        information during processing
    add_closure : bool
        If True, adds a final point to the knot near to the start point,
        so that it will appear visually to close when plotted.
    zero_centroid : bool
        If True, shifts the coordinates of the points so that their centre
        of mass is at the origin.
    '''

    def __init__(self, points, verbose=True, add_closure=False,
                 zero_centroid=False):
        if isinstance(points, SpaceCurve):
            points = points.points.copy()
        self._points = n.zeros((0, 3))
        self._crossings = None  # Will store a list of crossings if
                                # self.crossings() has been called
        self.points = n.array(points).astype(n.float)
        self.verbose = verbose

        self._cached_writhe_and_crossing_numbers = None
        self._gauss_code = None
        self._representation = None

        self._recent_octree = None

        if add_closure:
            self._add_closure()
        if zero_centroid:
            self.zero_centroid()

    def copy(self):
        '''Returns another knot with the same points and verbosity
        as self. Other attributes (e.g. cached crossings) are not
        preserved.'''
        return SpaceCurve(self.points.copy(), verbose=self.verbose)

    def tangents(self, closed=False):
        ts = np.roll(self.points, -1, axis=0) - self.points
        if not closed:
            ts = ts[:-1]
        return ts

    @property
    def points(self):
        '''
        The points of the spacecurve, as an Nx3 numpy array.
        '''
        return self._points

    @points.setter
    def points(self, points):
        self._points = points
        # Changing the points may change the crossings, so
        # reset any cached crossing attributes.
        self._crossings = None
        self._cached_writhe_and_crossing_numbers = None
        self._gauss_code = None

    def roll(self, distance, fix_endpoints=True):
        # if fix_endpoints, adds a point at the end that matches the
        # point at the start
        points = self.points
        if mag(points[-1] - points[0]) < 0.000001 and fix_endpoints:
            points = self.points[:-1]
        points = np.roll(points, distance, axis=0)
        if fix_endpoints:
            points = np.vstack([points, points[:1]])
        self.points = points

    def _vprint(self, s, newline=True):
        '''Prints s, with optional newline. Intended for internal use
        in displaying progress.'''
        if self.verbose:
            sys.stdout.write(s)
            if newline:
                sys.stdout.write('\n')
            sys.stdout.flush()

    def _add_closure(self):
        closing_distance = mag(self.points[-1] - self.points[0])
        if closing_distance > 0.02:
            self.points = n.vstack((self.points, self.points[:1] -
                                    0.001*(self.points[0] - self.points[-1])))

    def _unwrap_periodicity(self, shape):
        '''
        Walks along the points of self, assuming that a periodic
        boundary on a
        lattice bounded by :param:`shape` has been crossed whenever one
        point is too far from the previous one. When this occurs,
        subtracts the lattice vector in this direction.

        Parameters
        ----------
        shape : array-like
            The x, y, z distances of the periodic boundary.
        '''

        dx, dy, dz = shape
        points = self.points
        for i in range(1, len(points)):
            prevLine = points[i-1]
            curLine = points[i]
            rest = points[i:]
            change = curLine - prevLine
            if -1.05*dx < change[0] < -0.95*dx:
                rest[:, 0] += dx
            if 1.05*dx > change[0] > 0.95*dx:
                rest[:, 0] -= dx
            if -1.05*dy < change[1] < -0.95*dy:
                rest[:, 1] += dy
            if 1.05*dy > change[1] > 0.95*dy:
                rest[:, 1] -= dy
            if -1.05*dz < change[2] < -0.95*dz:
                rest[:, 2] += dz
            if 1.05*dz > change[2] > 0.95*dz:
                rest[:, 2] -= dz

        self.points = points

    @classmethod
    def from_gauss_code(cls, code):
        '''Creates a Knot from the given code, which must be provided as a
        string and may optionally include crossing orientations (these are
        actually ignored).'''
        from pyknotid.representations import Representation
        if 'a' in code or 'c' in code:
            k = cls(Representation(code).space_curve())
        else:
            k = cls(Representation.calculating_orientations(code).space_curve())
        k.rotate((0.05, 0.03, 0.02))
        return k

    @classmethod
    def closing_on_sphere(cls, line, com=(0., 0., 0.)):
        '''Adds new vertices to close the line at its maximum radius,
        returning a SpaceCurve representing the result.

        Parameters
        ----------
        line : ndarray
            The points of the line.
        com : iterable
            Optional additional centre of mass to shift by before closing

        '''
        com = ensure_shape_tuple(com)

        points = line - n.array(com)

        start = points[0]
        end = points[-1]

        start_r = mag(start)
        end_r = mag(end)

        start_theta = n.arccos(start[2] / start_r)
        end_theta = n.arccos(end[2] / end_r)

        start_phi = n.arctan2(start[1], start[0])
        end_phi = n.arctan2(end[1], end[0])

        rs = n.linspace(end_r, start_r, 100)
        thetas = n.linspace(end_theta, start_theta, 100)
        phis = n.linspace(end_phi, start_phi, 100)

        join_points = n.zeros((100, 3))
        join_points[:, 2] = rs * n.cos(thetas)
        join_points[:, 0] = rs * n.sin(thetas) * n.cos(phis)
        join_points[:, 1] = rs * n.sin(thetas) * n.sin(phis)

        return cls(n.vstack((points, join_points[1:-1])) + n.array(com))

    @classmethod
    def from_periodic_line(cls, line, shape, perturb=True, **kwargs):
        '''Returns a :class:`SpaceCurve` instance in which the line has been
        unwrapped through
        the periodic boundaries.

        Parameters
        ----------
        line : array-like
            The Nx3 vector of points in the line
        shape : array-like
            The x, y, z distances of the periodic boundary
        perturb : bool
            If True, translates and rotates the knot to avoid any lattice
            problems.
        '''
        shape = ensure_shape_tuple(shape)
        knot = cls(line, **kwargs)
        knot._unwrap_periodicity(shape)
        if perturb:
            knot.translate(n.array([0.00123, 0.00231, 0.00321]))
            knot.rotate((0.0002, 0.0001, 0.0001))
        return knot

    @classmethod
    def from_lattice_data(cls, line):
        '''Returns a :class:`SpaceCurve` instance in which the line has been
        slightly translated and rotated, in order to (practically) ensure
        no self intersections in closure or coincident points in
        projection.

        Parameters
        ----------
        line : array-like
            The list of points in the line. May be any type that SpaceCurve
            normally accepts.

        Returns
        -------
        :class:`SpaceCurve`
        '''
        knot = cls(line)
        knot.translate(n.array([0.00123, 0.00231, 0.00321]))
        knot.rotate(n.random.random(3) * 0.012)
        return knot

    @classmethod
    def from_braid_word(cls, word):
        '''
        Returns a :class:`SpaceCurve` instance formed from the given braid
        word.

        The braid word should be of the form 'aAbBcC' (i.e. capitalisation
        denotes inverse).

        Parameters
        ----------
        word : str
            The braid word to interpret.
        '''

        num_strands = len(set(word.lower())) + 1

        xs = n.arange(num_strands) + 1

        y = 0.

        lines = [[[x, y, 0.]] for x in xs]

        strands_by_final_index = {strand: lines[strand] for strand in range(num_strands)}
        strands_by_initial_index = {strand: lines[strand] for strand in range(num_strands)}

        for letter in word:
            index = ord(letter.lower()) - ord('a')

            sign = '+' if letter.isupper() else '-'

            if sign == '+':
                lines[index].append([index + 1 + 0.5, y + 0.5, 1.])
                lines[index].append([index + 1 + 1., y + 1., 0.])
                lines[index + 1].append([index + 1 + 0.5, y + 0.5, -1.])
                lines[index + 1].append([index + 1, y + 1., 0.])
            else:
                lines[index + 1].append([index + 1 + 0.5, y + 0.5, 1.])
                lines[index + 1].append([index + 1, y + 1., 0.])
                lines[index].append([index + 1 + 0.5, y + 0.5, -1.])
                lines[index].append([index + 1 + 1., y + 1., 0.])

            lines[index], lines[index + 1] = lines[index + 1], lines[index]
            strands_by_final_index[index], strands_by_final_index[index + 1] = strands_by_final_index[index + 1], strands_by_final_index[index]

            y += 1.

            for i, line in enumerate(lines):
                if i not in (index, index + 1):
                    lines[i].append([i + 1., y, 0.])

        for i, line in enumerate(lines):
            line.append([0., y + i + 1, 0.])
            line.append([-1*(i + 1), y, 0.])
            line.append([-1*(i + 1), 0, 0.])
            line.append([0., -i - 1., 0.])
            line.append([i + 1., -0.5, 0.])

        # indices_by_strand = {strand: index for (strand, index) in strands_by_final_index.items()}

        line = []
        index = 0
        for i in range(num_strands):
            cur_strand = strands_by_initial_index[index]
            line.extend(cur_strand)
            print('index is', index)
            print('cur strand is', cur_strand)
            index = int(np.round(cur_strand[-1][0])) - 1

        k = cls(n.array(line)*5.)
        k.zero_centroid()
        return k

    def scale(self, factor):
        '''Scales all the points of self by the given factor.

        You can accomplish the same thing, or other more subtle
        transformations, by modifying self.:py:attr:`points`.
        '''
        self.points *= factor

    def translate(self, vector):

        '''Translates all the points of self by the given vector.

        Parameters
        ----------
        vector : array-like
            The x, y, z translation distances
        '''
        self.points = self.points + n.array(vector)

    def zero_centroid(self):
        '''
        Translate such that the centroid (average position of vertices)
        is at (0, 0, 0).
        '''
        centroid = n.average(self.points, axis=0)
        self.translate(-1*centroid)

    def rotate(self, angles=None):
        '''
        Rotates all the points of self by the given angles in each axis.

        Parameters
        ----------
        angles : array-like
            The rotation angles about x, y and z. If None, random
            angles are used. Defaults to None.
        '''
        if angles is None:
            angles = n.random.random(3)
        phi, theta, psi = angles
        rot_mat = get_rotation_matrix(angles)
        self._apply_matrix(rot_mat)

    def _apply_matrix(self, mat):
        '''
        Applies the given matrix to all of self.points.
        '''
        self.points = n.apply_along_axis(mat.dot, 1, self.points)

    def cuaps(self, include_closure=True):
        '''Returns a list of the 'cuaps', where the curve is parallel to the
        positive x-axis. See D Bar-Natan and R van der Veen, "A
        polynomial time knot polynomial", 2017.
        '''
        points = self.points
        if mag(points[-1] - points[0]) < 0.000001:
            points = points[:-1]  # Avoids a badly defined tangent
        tangents = np.roll(points, -1, axis=0) - points
        angles = np.arctan2(tangents[:, 1], tangents[:, 0])

        print('angles', angles)

        cuaps = []

        for i in range(len(points) - 2):
            angle = angles[i]
            next_angle = angles[i + 1]

            if (0 < angle < np.pi / 2.) and (-np.pi / 2. < next_angle < 0):
                cuaps.append([i + 1, -1])
            elif (-np.pi / 2. < angle < 0) and (0 < next_angle < np.pi / 2.):
                cuaps.append([i + 1, 1])

        if include_closure:
            angle = angles[-2]
            next_angle = angles[-1]
            if (0 < angle < np.pi / 2.) and (-np.pi / 2. < next_angle < 0):
                cuaps.append([len(angles) - 1, -1])
            elif (-np.pi / 2. < angle < 0) and (0 < next_angle < np.pi / 2.):
                cuaps.append([len(angles) - 1, 1])

            angle = angles[-1]
            next_angle = angles[0]
            if (0 < angle < np.pi / 2.) and (-np.pi / 2. < next_angle < 0):
                cuaps.append([0, -1])
            elif (-np.pi / 2. < angle < 0) and (0 < next_angle < np.pi / 2.):
                cuaps.append([0, 1])

        cuaps = np.array(cuaps)
        # cuaps[:, -1] *= -1

        return cuaps

    def raw_crossings(self, mode='use_max_jump', include_closure=True,
                      recalculate=False, try_cython=True):

        '''Returns the crossings in the diagram of the projection of the
        space curve into its z=0 plane.

        The crossings will be calculated the first time this function
        is called, then cached until an operation that would change
        the list (e.g. rotation, or changing ``self.points``).

        Multiple modes are available (see parameters) - you should be
        aware of this because different modes may be vastly slower or
        faster depending on the type of line.

        Parameters
        ----------
        mode : str, optional
            One of ``'count_every_jump'`` or ``'use_max_jump'`` or
            ``'naive'``. In the first case,
            walking along the line uses information about the length of
            every step. In the second, it guesses that all steps have the
            same length as the maximum step length. In the last, no
            optimisation is made and every crossing is checked.
            The optimal choice depends on the data but is usually
            ``'use_max_jump'``, which is the default.
        include_closure : bool, optional
            Whether to include crossings with the
            line joining the start and end points. Defaults to True.
        recalculate : bool, optional
            Whether to force a recalculation of the crossing positions.
            Defaults to False.
        try_cython : bool, optional
            Whether to force the use of the python (as opposed to cython)
            implementation of find_crossings. This will make no difference
            if the cython could not be loaded, in which case python is already
            used automatically. Defaults to True.

        Returns
        -------
        array-like
            The raw array of floats representing crossings, of the
            form [[line_index, other_index, +-1, +-1], ...], where the
            line_index and other_index are in arclength parameterised
            by integers for each vertex and linearly interpolated,
            and the +-1 represent over/under and clockwise/anticlockwise
            respectively.
        '''

        if not recalculate and self._crossings is not None:
            return self._crossings

        if try_cython:
            helpers_module = chelpers
        else:
            helpers_module = helpers

        self._vprint('Finding crossings')

        points = self.points
        segment_lengths = n.roll(points[:, :2], -1, axis=0) - points[:, :2]
        segment_lengths = n.sqrt(n.sum(segment_lengths * segment_lengths,
                                       axis=1))
        # if include_closure:
        #     max_segment_length = n.max(segment_lengths)
        # else:
        max_segment_length = n.max(segment_lengths[:-1])

        numtries = len(points) - 3

        crossings = []

        jump_mode = {'count_every_jump': 1, 'use_max_jump': 2,
                     'naive': 3}[mode]

        for i in range(len(points)-2):
            if self.verbose:
                if i % 100 == 0:
                    self._vprint('\ri = {} / {}'.format(i, numtries),
                                 False)
            v0 = points[i]
            dv = points[(i+1) % len(points)] - v0

            s = points[i+2:]
            vnum = i
            compnum = i+2

            crossings.extend(helpers_module.find_crossings(
                v0, dv, s, segment_lengths[compnum:],
                vnum, compnum,
                max_segment_length,
                jump_mode
                ))

        if include_closure:
            closure_segment_length = segment_lengths[-1]
            max_segment_length = n.max([closure_segment_length,
                                        max_segment_length])
            v0 = points[-1]
            dv = points[0] - v0
            s = points[1:-1]
            vnum = len(points) - 1
            compnum = 1
            crossings.extend(helpers_module.find_crossings(
                v0, dv, s, segment_lengths[compnum:],
                vnum, compnum,
                max_segment_length,
                jump_mode))

        self._vprint('\n{} crossings found\n'.format(len(crossings) / 2))
        crossings.sort(key=lambda s: s[0])
        crossings = n.array(crossings)
        self._crossings = crossings

        return crossings

    def planar_writhe(self, **kwargs):
        '''
        Returns the current planar writhe of the knot; the signed sum
        of crossings of the current projection.

        The 'true' writhe is the average of this quantity over all
        projection directions, and is available from the :meth:`writhe`
        method.

        Parameters
        ----------
        **kwargs :
            These are passed directly to :meth:`raw_crossings`.
        '''
        crossings = self.raw_crossings(**kwargs)
        return n.sum(crossings[:, 3]) / 2.

    def writhe(self, samples=10, recalculate=False, method='integral', **kwargs):
        '''
        The (approximate) writhe of the space curve, obtained by averaging
        the planar writhe over the given number of directions.

        Parameters
        ----------
        samples : int
            The number of directions to average over. Defaults to 10.
        recalculate : bool
            Whether to recalculate the writhe even if a cached result
            is available. Defaults to False.
        method : str
            If 'projections', averages the planar writhe over many
            projections. If 'integral', calculates the writhing integral.
        **kwargs :
            These are passed directly to :meth:`raw_crossings`.
        '''
        if method == 'projections':
            crossing_number, writhe = self._writhe_and_crossing_numbers(
                samples, recalculate=recalculate, **kwargs)

        elif method == 'integral':
            from pyknotid.spacecurves.complexity import writhe_integral
            writhe = writhe_integral(self.points, **kwargs)

        return writhe

    def higher_order_writhe(self, order=(1, 3, 2, 4), try_cython=True):
        from pyknotid.spacecurves.complexity import higher_order_writhe_integral
        return higher_order_writhe_integral(self.points, order=order,
                                            try_cython=try_cython)

    def second_order_writhes(self, try_cython=True, **kwargs):
        from pyknotid.spacecurves.complexity import second_order_writhes
        return second_order_writhes(self.points, try_cython=try_cython, **kwargs)

    def planar_second_order_writhe(self, **kwargs):
        '''The second order writhe (type 2, i1,i3,i2,i4) of the projection of
        the curve along the z axis.
        '''
        from pyknotid.invariants import second_order_writhe
        gc = self.gauss_code(**kwargs)
        return second_order_writhe(gc)

    def second_order_twist(self, z):
        from pyknotid.spacecurves import complexity as com
        z = np.array(z).astype(np.float)
        assert len(z) == 3
        return com.second_order_twist(self.points, z)

    def average_crossing_number(self, samples=10, recalculate=False,
                                **kwargs):
        '''
        The (approximate) average crossing number of the space curve,
        obtained by averaging
        the planar writhe over the given number of directions.

        Parameters
        ----------
        samples : int
            The number of directions to average over.
        recalculate : bool
            Whether to recalculate the ACN.
        **kwargs :
            These are passed directly to :meth:`raw_crossings`.
        '''
        crossing_number, writhe = self._writhe_and_crossing_numbers(
            samples, recalculate=recalculate, **kwargs)
        return crossing_number

    def _writhe_and_crossing_numbers(self, samples=10, recalculate=False,
                                     include_closure=False,
                                     **kwargs):
        '''
        Calculates and stores the writhe and average crossing number.
        Internal (not intended for external use).
        '''
        if (self._cached_writhe_and_crossing_numbers is not None and
            self._cached_writhe_and_crossing_numbers[0] == samples and
            not recalculate):
            return self._cached_writhe_and_crossing_numbers[1]

        from .complexity import writhe_and_crossing_number
        numbers = writhe_and_crossing_number(self.points, samples,
                                             verbose=self.verbose,
                                             include_closure=include_closure,
                                             **kwargs)
        self._cached_writhe_and_crossing_numbers = (samples, numbers)

        return numbers

    def gauss_code(self, recalculate=False, **kwargs):
        '''
        Returns a :class:`~pyknotid.representations.gausscode.GaussCode`
        instance representing the crossings of the knot.

        The GaussCode instance is cached internally. If you want to
        recalculate it (e.g. to get an unsimplified version if you
        have simplified it), you should pass `recalculate=True`.

        This method passes kwargs directly to :meth:`raw_crossings`,
        see the documentation of that function for all options.
        '''

        from ..representations.gausscode import GaussCode
        if not recalculate and self._gauss_code is not None:
            return self._gauss_code
        crossings = self.raw_crossings(recalculate=recalculate, **kwargs)
        gc = GaussCode(crossings, verbose=self.verbose)
        self._gauss_code = gc
        return gc

    def representation(self, recalculate=False, **kwargs):
        '''
        Returns a :class:`~pyknotid.representations.representation.Representation`
        instance representing the crossings of the knot.

        The Representation instance is cached internally. If you want to
        recalculate it (e.g. to get an unsimplified version if you
        have simplified it), you should pass `recalculate=True`.

        This method passes kwargs directly to :meth:`raw_crossings`,
        see the documentation of that function for all options.
        '''

        from ..representations.representation import Representation
        if not recalculate and self._representation is not None:
            return self._representation
        crossings = self.raw_crossings(recalculate=recalculate, **kwargs)
        gc = Representation(crossings, verbose=self.verbose)
        self._representation = gc
        return gc
        

    def planar_diagram(self, **kwargs):
        '''
        Returns a
        :class:`~pyknotid.representations.planardiagram.PlanarDiagram`
        instance representing the crossings of the knot.

        This method passes kwargs directly to :meth:`raw_crossings`,
        see the documentation of that function for all options.
        '''
        from ..representations.planardiagram import PlanarDiagram
        crossings = self.raw_crossings(**kwargs)
        return PlanarDiagram(crossings)

    def gauss_diagram(self, simplify=False, **kwargs):
        '''
        Returns a
        :class:`~pyknotid.representations.gaussdiagram.GaussDiagram`
        instance representing the crossings of the knot.

        This method passes kwargs directly to :meth:`raw_crossings`,
        see the documentation of that function for all options.
        '''
        from ..representations.gaussdiagram import GaussDiagram
        gc = self.gauss_code(**kwargs)
        if simplify:
            gc.simplify()
        return GaussDiagram(gc)

    def reconstructed_space_curve(self):
        r = self.representation()
        return SpaceCurve(r.space_curve())

    def plot(self, mode='auto', clf=True, closed=False,
             **kwargs):
        '''
        Plots the line. See :func:`pyknotid.visualise.plot_line` for
        full documentation.
        '''
        plot_line(self.points, mode=mode, clf=clf,
                  closed=closed, **kwargs)

    def plot_projection(self, with_crossings=True, mark_start=False,
                        fig_ax=None, show=True, mark_points=False):
        '''Plots a 2D diagram of the knot projected along the current
        z-axis. The crossings, and start point of the curve, can
        optionally be marked.

        The projection is drawn using matplotlib.

        .. image:: example_knot_projection_9_5.png
           :align: center
           :scale: 50%
           :alt: An example knot projection.

        Parameters
        ----------
        with_crossings : bool
            If True, marks the location of each self-intersection in
            projection. Defaults to True.
        mark_start : bool
            If True, marks the first point of the curve. Default to False.
        fig_ax : tuple
            A 2-tuple of the matplotlib (fig, ax) to use, or None
            to create a new pair.
        show : bool
            If True, opens a new window showing the drawing. Defaults
            to True.

        Returns
        -------
        A 2-tuple of the matplotlib (fig, ax) used for the drawing.
        '''
        points = self.points
        crossings = None
        plot_crossings = []
        if with_crossings:
            crossings = self.raw_crossings()
            plot_crossings = []
            for crossing in crossings:
                x, y, over, orientation = crossing
                xint = int(x)
                r = points[xint]
                dr = points[(xint+1) % len(points)] - r
                plot_crossings.append(r + (x-xint) * dr)
        fig, ax = plot_projection(points,
                                  crossings=n.array(plot_crossings),
                                  mark_start=mark_start,
                                  mark_points=mark_points,
                                  fig_ax=fig_ax,
                                  show=show)
        return fig, ax

    def __str__(self):
        if self._crossings is not None:
            return '<SpaceCurve with {} points, {} crossings>'.format(
                len(self.points), len(self._crossings))
        return '<SpaceCurve with {} points>'.format(len(self.points))

    def __repr__(self):
        return str(self)

    def to_json(self, filen):
        '''
        Writes the knot points to the given filename, in a json format
        that can be read later by :meth:`SpaceCurve.from_json`. Uses
        :func:`pyknotid.io.to_json_file` internally.
        '''
        to_json_file(self.points, filen)

    @classmethod
    def from_json(cls, filen):
        '''
        Loads knot points from the given filename, assuming json format,
        and returns a :class:`SpaceCurve` with those points.
        '''

        points = from_json_file(filen)
        return cls(points)

    @classmethod
    def from_csv(cls, filen, **kwargs):
        '''
        Loads knot points from the given csv file, and returns a
        :class:`SpaceCurve` with those points.
        
        Arguments are passed straight to :func:`pyknot.io.from_csv`.
        '''
        return cls(from_csv(filen, **kwargs))

    def to_txt(self, filen):
        '''
        Writes the knot points to the given filename, formatted
        with each x,y,z component of each point space-separated
        on its own line, i.e.::

            ...
            1.2 6.1 98.5
            6.19 8.5 1.9
            ...
        '''
        with open(filen, 'w') as fileh:
            for line in self.points:
                fileh.write('{} {} {}\n'.format(line[0], line[1], line[2]))

    def octree_simplify(self, runs=1, plot=False, rotate=True,
                        obey_knotting=True, **kwargs):
        '''
        Simplifies the curve via the octree reduction of
        :module:`pyknotid.simplify.octree`.

        Parameters
        ----------
        runs : int
            The number of times to run the octree simplification.
            Defaults to 1.
        plot : bool
            Whether to plot the curve after each run. Defaults to False.
        rotate : bool
            Whether to rotate the space curve before each run. Defaults
            to True as this can make things much faster.
        obey_knotting : bool
            Whether to not let the line pass through itself. Defaults to
            True as this is always what you want for a closed curve.
        **kwargs
            Any remaining kwargs are passed to the
            :class:`pyknotid.simplify.octree.OctreeCell`
            constructor.

        '''
        from pyknotid.simplify.octree import OctreeCell
        for i in range(runs):
            if len(self.points) > 30:
                self._vprint('\rRun {} of {}, {} points remain'.format(
                    i, runs, len(self.points)))

            if rotate:
                rot_mat = get_rotation_matrix(n.random.random(3))
                self._apply_matrix(rot_mat)

            oc = OctreeCell.from_single_line(self.points, **kwargs)
            oc.simplify(obey_knotting)
            self._recent_octree = oc
            self.points = oc.get_single_line()

            if rotate:
                self._apply_matrix(rot_mat.T)

            if plot:
                self.plot()

        self._vprint('\nReduced to {} points'.format(len(self.points)))

    def arclength(self, include_closure=True):
        '''
        Returns the arclength of the line, the sum of lengths of each
        piecewise linear segment.

        Parameters
        ----------
        include_closure : bool
            Whether to include the distance between the final and
            first points. Defaults to True.
        '''
        return arclength(self.points, include_closure)

    def radius_of_gyration(self):
        '''
        Returns the radius of gyration of the points of self,
        assuming each has equal weight and ignoring the connecting
        lines.
        '''
        return radius_of_gyration(self.points)

    def __len__(self):
        return len(self.points)

    def reparameterised(self, mode='arclength', num_points=None,
                        interpolation='linear'):
        '''
        Returns a new :class:`SpaceCurve` where new points have been selected
        by interpolating the current ones.

        .. warning:: This doesn't do what you expect! The new segments
                     will probably not all be separated by the right amount
                     in terms of the new parameterisation.

        Parameters
        ----------
        mode : str
            The function to reparameterise by. Defaults to 'arclength',
            which is currently the only option.
        num_points : int
            The number of points in the new parameterisation. Defaults
            to None, which means the same as the current number.
        interpolation : str
            The type of interpolation to use, passed directly to the
            ``kind`` option of ``scipy.interpolate.interp1d``. Defaults
            to 'linear', and other options have not been tested.
        '''
        from scipy.interpolate import interp1d
        indices = self._new_indices_by_arclength(num_points)

        interp_xs = interp1d(range(len(self.points)+1),
                             n.hstack((self.points[:, 0],
                                      self.points[:, 0][:1])))
        interp_ys = interp1d(range(len(self.points)+1),
                             n.hstack((self.points[:, 1],
                                      self.points[:, 1][:1])))
        interp_zs = interp1d(range(len(self.points)+1),
                             n.hstack((self.points[:, 2],
                                      self.points[:, 2][:1])))

        new_points = n.zeros((len(indices), 3), dtype=n.float)
        new_points[:, 0] = interp_xs(indices)
        new_points[:, 1] = interp_ys(indices)
        new_points[:, 2] = interp_zs(indices)

        return SpaceCurve(new_points)

    def _new_indices_by_arclength(self, number, step=None, gap=0):
        if number is None:
            number = len(self.points)
        total_arclength = self.arclength()
        if step is None:
            arclengths = n.linspace(0, total_arclength - gap, number+1)[:-1]
        else:
            arclengths = n.arange(0, total_arclength - gap, step)

        arclengths[0] += 0.000001

        points = self.points
        segment_arclengths = self.segment_arclengths()
        cumulative_arclength = n.hstack([[0.], n.cumsum(segment_arclengths)])
        total_arclength = self.arclength()

        indices = []

        for arclength in arclengths:
            first_greater_index = n.argmax(cumulative_arclength > arclength)
            last_lower_index = (first_greater_index - 1) % len(points)
            arclength_below = cumulative_arclength[last_lower_index]
            step_arclength = segment_arclengths[last_lower_index]
            step_fraction = (arclength - arclength_below) / step_arclength
            indices.append(last_lower_index + step_fraction)

        return indices

    def segment_arclengths(self):
        '''
        Returns an array of arclengths of every step in the line
        defined by self.points.
        '''
        return n.apply_along_axis(
            mag, 1, n.roll(self.points, -1, axis=0) - self.points)

    def smooth(self, repeats=1, periodic=True, window_len=10,
               window='hanning'):
        '''Smooths each of the x, y and z components of self.points by
        convolving with a window of the given type and size.

        .. warning:: This is *not* topologically safe, it can change
                     the knot type of the curve. For topologically
                     safe reduction, see :meth:`octree_simplify`.

        Parameters
        ----------
        repeats : int
            Number of times to run the smoothing algorithm. Defaults to 1.
        periodic : bool
            If True, the convolution window wraps around the curve.
            Defaults to True.
        window_len : int
            Width of the convolution window. Defaults to 10.
            Passed to :func:`pyknotid.spacecurves.smooth.smooth`.
        window : string
            The type of convolution window. Defaults to 'hanning'.
            Passed to :func:`pyknotid.spacecurves.smooth.smooth`.
        '''
        points = self.points
        if periodic:
            points = n.vstack((points[-(window_len + 1):],
                               points,
                               points[:(window_len + 1)]))
        else:
            start = points[0]
            end = points[-1]
        if len(points) > window_len:
            for i in range(repeats):
                points[:, 0] = smooth(points[:, 0], window_len, window)
                points[:, 1] = smooth(points[:, 1], window_len, window)
                points[:, 2] = smooth(points[:, 2], window_len, window)
                if not periodic:
                    points[0] = start
                    points[-1] = end
        if periodic:
            points = points[(window_len + 1):-(window_len + 1)]
        self.points = points

    def simplify_straight_segments(self, closed=False):
        '''Replaces successive curve segments with identical tangents with a
        single longer segment.'''

        points = self.points
        ts = np.roll(points, -1, axis=0) - points

        keep_points = np.ones(len(points), dtype=np.bool)

        ts_range = ts[:-2] if not closed else ts

        for i1, t1 in enumerate(ts_range):
            t1 /= np.sqrt(np.sum(t1**2))
            t2 = ts[(i1 + 1) % len(ts)]
            t2 /= np.sqrt(np.sum(t2**2))
            if np.dot(t1, t2) > (1 - 10e-10):
                keep_points[i1 + 1] = 0

        self.points = self.points[keep_points]

    def curvatures(self, closed=True):
        '''Returns curvatures at each vertex (or really line segment).'''
        points = self.points

        ts = np.roll(points, -1, axis=0) - points
        vs = np.sqrt(np.sum(ts*ts, axis=1))

        gamma_dd = 0.25 * (np.roll(points, -2, axis=0) + np.roll(points, 2, axis=0) - 2*points)
        vns = 0.5 * (vs + np.roll(vs, 1))
        gamma_d = 0.5 * (np.roll(points, -1, axis=0) - np.roll(points, 1, axis=0))
        ans = 0.25 * (np.roll(vs, -1) + vs - np.roll(vs, 1) - np.roll(vs, 2))
        kappas = gamma_dd - np.einsum('i,ij->ij', (ans / vns), gamma_d)
        kappas = np.sqrt(np.sum(kappas * kappas, axis=1))
        kappas = (1./vns**2) * kappas

        if not closed:
            kappas[:3] = kappas[3]
            kappas[-3:] = kappas[-4]

        return kappas

    def torsions(self, signed=False, closed=True):
        '''Returns torsions at each vertex.'''

        d = self.points

        ts = np.roll(d, -1, axis=0) - d
        vs = np.sqrt(np.sum(ts*ts, axis=1))

        gamma_dd = 0.25 * (np.roll(d, -2, axis=0) + np.roll(d, 2, axis=0) - 2*d)
        vns = 0.5 * (vs + np.roll(vs, 1))
        gamma_d = 0.5 * (np.roll(d, -1, axis=0) - np.roll(d, 1, axis=0))
        ans = 0.25 * (np.roll(vs, -1) + vs - np.roll(vs, 1) - np.roll(vs, 2))

        kappas = gamma_dd - np.einsum('i,ij->ij', (ans / vns), gamma_d)
        kappas = np.sum(kappas * kappas, axis=1)

        gamma_ddd = 0.125 * (np.roll(d, -3, axis=0) - 3*np.roll(d, -1, axis=0) +
                            3*np.roll(d, 1, axis=0) - np.roll(d, 3, axis=0))

        numerator = np.sum(np.cross(gamma_d, gamma_dd) * gamma_ddd, axis=1)
        denominator = vns**2 * np.abs(kappas)

        torsions = numerator / denominator
        if not signed:
            torsions = np.abs(torsions)

        if not closed:
            torsions[:3] = torsions[3]
            torsions[-3:] = torsions[-4]

        return torsions

    def close(self):
        '''Adds the starting point to the end of the curve, so that it ends
        exactly where it began.

        '''
        self.points = np.vstack([self.points, self.points[:1]])

    def interpolate(self, num_points, s=0, **kwargs):
        '''Replaces self.points with points from a B-spline interpolation.

        This method uses scipy.interpolate.splprep. kwargs are passed
        to this function.
        '''
        from scipy.interpolate import splprep, splev

        lengths = self.segment_arclengths()
        cum_length_fracs = np.cumsum(lengths) / np.sum(lengths)

        tck, u = splprep([self.points[:, 0],
                          self.points[:, 1],
                          self.points[:, 2]],
                         u=cum_length_fracs,
                         **kwargs)
        points = splev(np.linspace(0, 1, num_points),
                            tck)

        new_points = np.zeros((num_points, 3))
        new_points[:, 0] = points[0]
        new_points[:, 1] = points[1]
        new_points[:, 2] = points[2]

        self.points = new_points






'''
Cython functions for octree calculations.
'''


import numpy as n
cimport numpy as n

cimport cython

from libc.math cimport abs, pow, sqrt as csqrt, floor, acos

cpdef angle_exceeds(double [:, :] ps, double val=2*n.pi,
                    long include_closure=1):
    '''Returns True if the sum of angles along ps exceeds
    val, else False.

    If include_closure, includes the angles with the line closing
    the end and start points.
    '''
    cdef double angle = 0.
    cdef double [:] nex = ps[0]
    cdef double [:] nex2 = ps[1]
    cdef double [:] dv2 = n.zeros(3, dtype=n.double)
    diff(dv2, nex, nex2)
    divide(dv2, mag(dv2))
    cdef double [:] cur
    cdef double increment
    cdef long lenps = len(ps)
    cdef long [:] checks = n.arange(len(ps)) if include_closure else n.arange(len(ps)-2)
    cdef int i
    for i in checks:
        cur = nex
        nex = nex2
        nex2 = ps[(i+2) % lenps]
        dv = dv2
        diff(dv2, nex, nex2)
        divide(dv2, mag(dv2))
        increment = angle_between(dv, dv2)
        if n.isnan(increment):
            return True
        angle += increment
        if angle > val:
            return True
    assert not n.isnan(angle)
    return False

cdef void diff(double [:] dv2, double [:] nex, double [:] nex2):
    dv2[0] = nex2[0] - nex[0]
    dv2[1] = nex2[1] - nex[1]
    dv2[2] = nex2[2] - nex[2]

cdef double angle_between(double [:] v1, double [:] v2):
    '''Returns angle between v1 and v2, assuming they are normalised to 1.'''
    # clip becaus v1.dot(v2) may exceed 1 due to floating point
    cdef double value = v1[0] * v2[0] + v1[1] * v2[1] + v1[2] * v2[2]
    if value > 1.:
        value = 1.
    elif value < 0.:
        value = 0.
    return value

cdef void divide(double [:] arr, double val):
    arr[0] = arr[0] / val
    arr[1] = arr[1] / val
    arr[2] = arr[2] / val

cdef void multiply(double [:] arr, double val):
    arr[0] = arr[0] * val
    arr[1] = arr[1] * val
    arr[2] = arr[2] * val

cdef double mag(double [:] v):
    return pow(v[0], 2) + pow(v[1], 2) + pow(v[2], 2)


#@cython.boundscheck(False)
#@cython.wraparound(False)
cpdef line_to_segments(line, cuts=None, join_ends=True):
    '''Takes a line (set of points), a list of cut planes in
    x, y, z, and a parameter to decide whether the line
    joining the first and last point should also be cut.

    Returns a list of shorter lines resulting from cutting at
    all these cut planes.'''

    cdef double [:, :] cy_line = line

    cdef double cut_x, cut_y, cut_z
    if cuts is None:
        xmin, ymin, zmin = n.min(line, axis=0) - 1
        xmax, ymax, zmax = n.max(line, axis=0) + 1
        cut_x = (xmax + xmin) / 2.
        cut_y = (ymax + ymin) / 2.
        cut_z = (zmin + zmax) / 2.
    else:
        cut_x, cut_y, cut_z = cuts


    cdef double [:] cy_dv = n.zeros(3, dtype=n.double)
    cdef double [:] cy_nex
    cdef double [:] cy_cur
    cdef double dx, dy, dz
    cdef double x_cut_pos, y_cut_pos, z_cut_pos
        
    # Cut the line wherever it passes through a quad cell boundary
    cdef list segments = []
    cdef long cut_i = 0
    cdef long i
    for i in range(len(line)-1):
        cy_cur = cy_line[i]
        cy_nex = cy_line[i+1]

        diff(cy_dv, cy_cur, cy_nex)
        dx = cy_dv[0]
        dy = cy_dv[1]
        dz = cy_dv[2]

        cross_cut_x = sign(cy_cur[0] - cut_x) != sign(cy_nex[0] - cut_x)
        cross_cut_y = sign(cy_cur[1] - cut_y) != sign(cy_nex[1] - cut_y)
        cross_cut_z = sign(cy_cur[2] - cut_z) != sign(cy_nex[2] - cut_z)

        if (not cross_cut_x and not cross_cut_y and not cross_cut_z):
            continue

        cur = line[i]
        nex = line[i+1]
        dv = cur - nex
        cur = line[i]
        nex = line[i+1]

        if cross_cut_x and cross_cut_y and cross_cut_z:
            x_cut_pos = -1 * (cur[0]-cut_x)/dx
            y_cut_pos = -1 * (cur[1]-cut_y)/dy
            z_cut_pos = -1 * (cur[2]-cut_z)/dz
            order = n.sort((x_cut_pos, y_cut_pos, z_cut_pos))
            # assert 0 < x_cut_pos < 1 and 0 < y_cut_pos < 1 and 0 < z_cut_pos < 1
            join_point_1 = cur + order[0]*dv
            join_point_2 = cur + order[1]*dv
            join_point_3 = cur + order[2]*dv
            first_seg = n.vstack((line[cut_i:(i+1)].copy(), join_point_1))
            second_seg = n.vstack((join_point_1, join_point_2))
            third_seg = n.vstack((join_point_2, join_point_3))
            line[i] = join_point_3
            cut_i = i
            segments.append(first_seg)
            segments.append(second_seg)
            segments.append(third_seg)
        elif cross_cut_x and cross_cut_y:
            x_cut_pos = -1 * (cur[0]-cut_x)/dx
            y_cut_pos = -1 * (cur[1]-cut_y)/dy
            order = n.sort((x_cut_pos, y_cut_pos))
            join_point_1 = cur + order[0]*dv
            join_point_2 = cur + order[1]*dv
            first_seg = n.vstack((line[cut_i:(i+1)].copy(), join_point_1))
            second_seg = n.vstack((join_point_1, join_point_2))
            line[i] = join_point_2
            cut_i = i
            segments.append(first_seg)
            segments.append(second_seg)
        elif cross_cut_x and cross_cut_z:
            x_cut_pos = -1 * (cur[0]-cut_x)/dx
            z_cut_pos = -1 * (cur[2]-cut_z)/dz
            order = n.sort((x_cut_pos, z_cut_pos))
            join_point_1 = cur + order[0]*dv
            join_point_2 = cur + order[1]*dv
            first_seg = n.vstack((line[cut_i:(i+1)].copy(), join_point_1))
            second_seg = n.vstack((join_point_1, join_point_2))
            line[i] = join_point_2
            cut_i = i
            segments.append(first_seg)
            segments.append(second_seg)
        elif cross_cut_y and cross_cut_z:
            y_cut_pos = -1 * (cur[1]-cut_y)/dy
            z_cut_pos = -1 * (cur[2]-cut_z)/dz
            order = n.sort((y_cut_pos, z_cut_pos))
            join_point_1 = cur + order[0]*dv
            join_point_2 = cur + order[1]*dv
            first_seg = n.vstack((line[cut_i:(i+1)].copy(), join_point_1))
            second_seg = n.vstack((join_point_1, join_point_2))
            line[i] = join_point_2
            cut_i = i
            segments.append(first_seg)
            segments.append(second_seg)
        elif cross_cut_x:
            cut_pos = -1 * (cur[0]-cut_x)/dx
            assert 0. <= cut_pos <= 1.
            join_point = cur + cut_pos*dv
            first_seg = n.vstack((line[cut_i:(i+1)].copy(), join_point))
            line[i] = join_point
            cut_i = i
            segments.append(first_seg)
        elif cross_cut_y:
            cut_pos = -1 * (cur[1]-cut_y)/dy
            assert 0. <= cut_pos <= 1.
            join_point = cur + cut_pos*dv
            first_seg = n.vstack((line[cut_i:(i+1)].copy(), join_point))
            line[i] = join_point
            cut_i = i
            segments.append(first_seg)
        elif cross_cut_z:
            cut_pos = -1 * (cur[2]-cut_z)/dz
            assert 0. <= cut_pos <= 1.
            join_point = cur + cut_pos*dv

            first_seg = n.vstack((line[cut_i:(i+1)].copy(), join_point))
            line[i] = join_point
            # second_seg = n.vstack((join_point, line[(i+1):]))

            cut_i = i
            segments.append(first_seg)

    final_seg = line[cut_i:]
    if cut_i > 0:
        if join_ends:
            first_seg = segments.pop(0)
            segments.append(n.vstack((final_seg, first_seg)))
        else:
            segments.append(final_seg)
    else:
        segments.append(final_seg)

    return segments



cdef double sign(double v):
   if v > 0:
       return 1.0
   elif v < 0:
       return -1.0
   return 0.0
   



'''
Random polygons from the quaternionic viewpoint [1]
===================================================

[1] These polygons are generated following the algorithms of (Cantarella, Deguchi and Shonkwiler. "Probability Theory of Random Polygons from the Quaternionic Viewpoint". Communications on Pure and Applied Mathematics 67 (2014).)
'''

import numpy as n

def get_closed_loop(length, seed=0, normalisation=7.5):
    '''
    Returns the points of a closed Cantarella walk with the given
    number of points.

    Parameters
    ----------
    length : int
        The number of segments in the walk.
    seed : int
        The random seed to use. Defaults to 0, which causes a new
        seed to be chosen randomly.
    normalisation : float
        The average segment length to normalise to. Defaults to 7.5,
        purely to work well with plotting defaults (tube radius 1).

    '''
    edges = closed_loop_segments(length, seed)
    return edges_to_path(edges) * normalisation * length / 2.

def get_open_line(length, seed=0, normalisation=7.5):
    '''
    Returns the points of an open Cantarella walk with the given
    number of points.

    Parameters
    ----------
    length : int
        The number of segments in the walk.
    seed : int
        The random seed to use. Defaults to 0, which causes a new
        seed to be chosen randomly.
    normalisation : float
        The average segment length to normalise to. Defaults to 7.5,
        purely to work well with plotting defaults (tube radius 1).
    '''
    edges = open_line_segments(length, seed)
    return edges_to_path(edges) * normalisation * length / 2.

def get_open_by_distance_line(length, distance=0., seed=0, normalisation=7.5):
    edges = open_by_distance_segments(length, distance, seed)
    return edges_to_path(edges) * normalisation * length / 2.

def open_line_segments(num, seed=0):
    if seed == 0:
        seed = n.random.randint(1000000)
        
    generator = n.random.RandomState()
    generator.seed(seed)
    
    # wjs is a vector uniformly distributed on the (4*num)-sphere
    # Such a vector is given by (4*num) normally distributed numbers in a
    # normalised vector
    wjs = generator.normal(size=4 * num)
    normfac = 1./n.sqrt(n.sum(wjs*wjs))
    wjs *= normfac
    
    # Redistribute wjs as sets of 4-points
    points = wjs.reshape((num, 4))
    
    edges = n.zeros((num, 3))
    
    # Create each edge as a transformation of the 4-points
    edges[:, 0] = (points[:, 0]**2 + points[:, 1]**2 - points[:, 2]**2 -
                   points[:, 3]**2)
    edges[:, 1] = 2 * (points[:, 1] * points[:, 2] - points[:, 0] *
                       points[:, 3])
    edges[:, 2] = 2 * (points[:, 0] * points[:, 2] - points[:, 1] *
                       points[:, 3])
    return edges


def closed_loop_segments(num, seed=0):
    if seed == 0:
        seed = n.random.randint(1000000)
        
    generator = n.random.RandomState()
    generator.seed(seed)
    
    uu = generator.normal(size=num) + 1j*generator.normal(size=num)
    vv = generator.normal(size=num) + 1j*generator.normal(size=num)
    
    u = uu / n.sqrt(uu.dot(uu.conj()))
    uconj = u.conj()
    vc = vv - u.conj().dot(vv)*u
    v = vc / n.sqrt(vc.dot(vc.conj()))
    vconj = v.conj()

    edges = n.zeros((num, 3),dtype=float)
    
    edges[:, 0] = (u * uconj - v * vconj).real
    edges[:, 1] = (-1j * (u * vconj - v * uconj)).real
    edges[:, 2] = (u * vconj + v * uconj).real
    
    return edges
    

def open_by_distance_segments(num, distance=0., seed=0):
    '''Get a list of line vectors whose sum fails to close by the given distance fraction.

    Parameters
    ----------
    num : int
        The number of segments
    distance : float
        The distance fraction (0 <= distance <= 1) by which their sum doesn't close.
    seed : int
        The random seed.
    '''
    if seed == 0:
        seed = n.random.randint(1000000)

    if distance > 1. or distance < 0.:
        raise ValueError('distance must be between 0 and 1')

    distance *= 2.

    generator = n.random.RandomState()
    generator.seed(seed)
    
    uu = generator.normal(size=num) + 1j*generator.normal(size=num)
    vv = generator.normal(size=num) + 1j*generator.normal(size=num)
    
    u = uu / n.sqrt(uu.dot(uu.conj()))
    uconj = u.conj()
    vc = vv - u.conj().dot(vv)*u
    v = vc / n.sqrt(vc.dot(vc.conj()))
    vconj = v.conj()

    # modu = n.sqrt(n.sum(u**2 + uconj**2))
    # print 'modu is', modu
    # u /= modu
    # uconj /= modu
    u *= n.sqrt(1 + distance / 2.)
    uconj *= n.sqrt(1 + distance / 2.)

    # modv = n.sqrt(n.sum(v**2 + vconj**2))
    # print 'modv is', modv
    # v /= modv
    # vconj /= modv
    v *= n.sqrt(1 - distance / 2.)
    vconj *= n.sqrt(1 - distance / 2.)

    # print 'mod u', n.sqrt(n.sum(u**2 + uconj**2)), n.sqrt(1 + distance/2.)
    # print 'mod v', n.sqrt(n.sum(v**2 + vconj**2)), n.sqrt(1 - distance/2.)

    edges = n.zeros((num, 3),dtype=float)
    
    edges[:, 0] = (u * uconj - v * vconj).real
    edges[:, 1] = (-1j * (u * vconj - v * uconj)).real
    edges[:, 2] = (u * vconj + v * uconj).real
    
    return edges

def edges_to_path(edges):
    num = len(edges)
    path = n.zeros((num + 1, 3))
    for i in range(num):
        path[i+1] = path[i] + edges[i]
    return path



'''
Representation
==============

An abstract representation of a Knot, providing methods for
the calculation of topological invariants.

API documentation
~~~~~~~~~~~~~~~~~
'''

from __future__ import print_function, division
from pyknotid.representations.gausscode import GaussCode
from collections import defaultdict
import numpy as n


class Representation(GaussCode):
    '''
    An abstract representation of a knot or link. Internally
    this is just a Gauss code, but it exposes extra topological methods
    and may in future be refactored to work differently.
    '''

    @classmethod
    def calculating_orientations(cls, code, **kwargs):
        gc = super(Representation, cls).calculating_orientations(code)
        return Representation(gc, **kwargs)

    def gauss_code(self):
        return GaussCode(self)
    
    def planar_diagram(self):
        from pyknotid.representations.planardiagram import PlanarDiagram
        return PlanarDiagram(self)

    def alexander_polynomial(self, variable=-1, quadrant='lr',
                             mode='python', force_no_simplify=False):
        '''
        Returns the Alexander polynomial at the given point,
        as calculated by :func:`pyknotid.invariants.alexander`.

        See :func:`pyknotid.invariants.alexander` for the meanings
        of the named arguments.
        '''
        from ..invariants import alexander
        if not force_no_simplify:
            self.simplify()
        return alexander(self, variable=variable, quadrant=quadrant,
                         simplify=False, mode=mode)

    def jones_polynomial(self, variable=-1, simplify=True):
        if simplify:
            self.simplify()
        from pyknotid.invariants import jones_polynomial
        p = self.planar_diagram()
        return jones_polynomial(p)

    def alexander_at_root(self, root, round=True, **kwargs):
        '''
        Returns the Alexander polynomial at the given root of unity,
        i.e. evaluated at exp(2 pi I / root).

        The result returned is the absolute value.

        Parameters
        ----------
        root : int
            The root of unity to use, i.e. evaluating at exp(2 pi I / root).
            If this is iterable, this method returns a list of the results
            at every value of that iterable.
        round : bool
            If True and n in (1, 2, 3, 4), the result will be rounded
            to the nearest integer for convenience, and returned as an
            integer type.
        **kwargs :
            These are passed directly to :meth:`alexander_polynomial`.
        '''
        if hasattr(root, '__contains__'):
            return [self.alexander_at_root(r) for r in root]
        variable = n.exp(2 * n.pi * 1.j / root)
        value = self.alexander_polynomial(variable, **kwargs)
        value = n.abs(value)
        if round and root in (1, 2, 3, 4):
            value = int(n.round(value))
        return value

    def vassiliev_degree_2(self, simplify=True):
        '''
        Returns the Vassiliev invariant of degree 2 for the Knot.

        Parameters
        ==========
        simplify : bool
            If True, simplifies the Gauss code of self before
            calculating the invariant. Defaults to True, but
            will work fine if you set it to False (and might even
            be faster).
        **kwargs :
            These are passed directly to :meth:`gauss_code`.
        '''
        from ..invariants import vassiliev_degree_2
        if simplify:
            self.simplify()
        return vassiliev_degree_2(self)

    def vassiliev_degree_3(self, simplify=True, try_cython=True):
        '''Returns the Vassiliev invariant of degree 3 for the Knot.

        Parameters
        ==========
        simplify : bool
            If True, simplifies the Gauss code of self before
            calculating the invariant. Defaults to True, but
            will work fine if you set it to False (and might even
            be faster).
        try_cython : bool
            Whether to try and use an optimised cython version of the
            routine (takes about 1/3 of the time for complex
            representations).  Defaults to True, but the python
            fallback will be *slower* than setting it to False if the
            cython function is not available.
        **kwargs :
            These are passed directly to :meth:`gauss_code`.

        '''
        from ..invariants import vassiliev_degree_3
        if simplify:
            self.simplify()
        return vassiliev_degree_3(self, try_cython=try_cython)

    def virtual_vassiliev_degree_3(self):
        '''Returns the virtual Vassiliev invariant of degree 3 for the
        representation.

        '''
        from ..invariants import virtual_vassiliev_degree_3
        return virtual_vassiliev_degree_3(self)

    def hyperbolic_volume(self):
        '''
        Returns the hyperbolic volume at the given point, via
        :meth:`pyknotid.representations.PlanarDiagram.as_spherogram`.
        '''
        from ..invariants import hyperbolic_volume
        return hyperbolic_volume(self.planar_diagram())

    def exterior_manifold(self):
        '''
        The knot complement manifold of self as a SnapPy class
        giving access to all of SnapPy's tools.

        This method requires that Spherogram, and possibly SnapPy,
        are installed.
        '''
        return self.planar_diagram().as_spherogram().exterior()

    def identify(self, determinant=True, vassiliev_2=True,
                 vassiliev_3=None, alexander=False, roots=(2, 3, 4),
                 min_crossings=True):
        '''
        Provides a simple interface to
        :func:`pyknotid.catalogue.identify.from_invariants`, by passing
        the given invariants. This does *not* support all invariants
        available, or more sophisticated identification methods,
        so don't be afraid to use the catalogue functions directly.

        Parameters
        ----------
        determinant : bool
            If True, uses the knot determinant in the identification.
            Defaults to True.
        alexander : bool
            If True-like, uses the full alexander polynomial in the
            identification. If the input is a dictionary of kwargs,
            these are passed straight to self.alexander_polynomial.
        roots : iterable
            A list of roots of unity at which to evaluate. Defaults
            to (2, 3, 4), the first of which is redundant with the
            determinant. Note that higher roots can be calculated, but
            aren't available in the database.
        min_crossings : bool
            If True, the output is restricted to knots with fewer crossings
            than the current projection of this one. Defaults to True. The
            only reason to turn this off is to see what other knots have
            the same invariants, it is never not useful for direct
            identification.
        vassiliev_2 : bool
            If True, uses the Vassiliev invariant of order 2. Defaults to True.
        vassiliev_3 : bool
            If True, uses the Vassiliev invariant of order 3. Defaults to None,
            which means the invariant is used only if the representation has
            less than 30 crossings.
        '''
        if not roots:
            roots = []
        roots = set(roots)
        if determinant:
            roots.add(2)

        if len(self) < 30 and vassiliev_3 is None:
            vassiliev_3 = True

        identify_kwargs = {}
        for root in roots:
            identify_kwargs[
                'alex_imag_{}'.format(root)] = self.alexander_at_root(root)

        if vassiliev_2:
            identify_kwargs['v2'] = self.vassiliev_degree_2()
        if vassiliev_3:
            identify_kwargs['v3'] = self.vassiliev_degree_3()

        if alexander:
            if not isinstance(alexander, dict):
                import sympy as sym
                alexander = {'variable': sym.var('t')}
            poly = self.alexander_polynomial(**alexander)
            identify_kwargs['alexander'] = poly

        if min_crossings and len(self.gauss_code()) < 16:
            identify_kwargs['max_crossings'] = len(self.gauss_code())

        from pyknotid.catalogue.identify import from_invariants
        return from_invariants(**identify_kwargs)

    def is_virtual(self):
        '''
        Takes an open curve and checks (for the default projection) if its 
        Gauss code corresponds to a virtual knot or not. Returns a Boolean of 
        this information.

        Returns
        -------
        virtual : bool
            True if the Gauss code corresponds to a virtual knot. False
            otherwise.
        '''
        if len(self._gauss_code) == 0:
            return False
        if len(self._gauss_code[0]) == 0:
            return False
        gauss_code = self._gauss_code[0][:, 0]
        l = len(gauss_code)
        total_crossings = l / 2
        crossing_counter = 1
        virtual = False
        
        for crossing_number in self.crossing_numbers:
            occurences = n.where(gauss_code == crossing_number)[0]
            first_occurence = occurences[0]
            second_occurence = occurences[1]
            crossing_difference = second_occurence - first_occurence        
                  
            if crossing_difference % 2 == 0:
                return True
        return False

    def self_linking(self):
        '''Returns the self linking number J(K) of the Gauss code, an
        invariant of virtual knots. See Kauffman 2004 for more information.

        Returns
        -------
        slink_counter : int
            The self linking number of the open curve
        '''
        from ..invariants import self_linking
        return self_linking(self)

    def writhe(self):
        writhe = 0

        return int(n.round(n.sum([n.sum(l[:, -1]) for l in self._gauss_code])/2.))

    def slip_triangle(self, func):

        code = self._gauss_code[0]

        length = len(self)
        array = n.ones((len(self) + 1, len(self) + 1)) * -0
        
        for i in range(length + 1):
            for j in range(length + 1):
                if i + j > length:
                    continue
                new_gc = Representation(self)

                for _ in range(i):
                    new_gc._remove_crossing(new_gc._gauss_code[0][0, 0])
                for _ in range(j):
                    new_gc._remove_crossing(new_gc._gauss_code[0][-1, 0])

                invariant = func(new_gc)

                array[-1*(i + 1), j] = invariant

        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        ax.imshow(array, interpolation='none', cmap='jet')

        ticks = range(length + 1)
        ax.set_xticks(ticks)
        ax.set_yticks(ticks)
        ax.set_xticklabels([str(t) for t in ticks])
        ax.set_yticklabels([str(t) for t in ticks])

        ax.plot([0, length+1], [0, length+1], color='black', linewidth=2)
        ax.set_xlim(-0.5, length+0.5)
        ax.set_ylim(length+0.5, -0.5)

        
        fig.show()

        return array, fig, ax

    def _construct_planar_graph(self):
        pd = self.planar_diagram()
        g, duplicates, heights, first_edge = pd.as_networkx_extended()

        import planarity

        pg = planarity.PGraph(g)
        pg.embed_drawplanar()
        g = planarity.networkx_graph(pg)


        node_labels = {}
        xs = []
        ys = []

        nodes_by_height = {}
        node_xs_by_y = {}
        node_xs_ys = {}
        node_lefts_rights = {}

        for node, data in g.nodes(data=True):
            y = data['pos']
            xb = data['start']
            xe = data['end']
            x = int((xe + xb) / 2.)

            node_labels[node] = (x, y)
            xs.extend([xb, xe])
            ys.append(y)

            nodes_by_height[data['pos']] = node
            node_xs_by_y[data['pos']] = x
            node_xs_ys[node] = (x, y)
            node_lefts_rights[node] = (xb, xe)

        lines = []

        rightmost_x = n.max(xs)
        leftmost_x = n.min(xs)
        x_span = rightmost_x - leftmost_x
        safe_yshift = 0.5 / x_span

        extra_x_shifts = []
        
        for n1, n2, data in g.edges(data=True):
            x = data['pos']
            yb = data['start']
            ye = data['end']

            start_node = nodes_by_height[yb]
            end_node = nodes_by_height[ye]
            if start_node >= len(self) and end_node >= len(self):
                continue

            start_left, start_right = node_lefts_rights[start_node]
            end_left, end_right = node_lefts_rights[end_node]

            start_frac = n.abs((x - start_left) / (start_right - start_left) - 0.5)
            start_frac = 0.5 - start_frac
            if True:  # ye < ys:  # This always evaluated to True - a bug?
                start_frac *= -1
            start_shift = start_frac

            end_frac = n.abs((x - end_left) / (end_right - end_left) - 0.5)
            end_frac = 0.5 - end_frac
            if False:  # ye > ys:  # This always evaluated to False - a bug?
                end_frac *= -1
            end_shift = end_frac

            start_node_x = node_xs_by_y[yb]
            start_node_y = yb
            
            end_node_x = node_xs_by_y[ye]
            end_node_y = ye

            line = n.array([[start_node_x, start_node_y],
                            [x, start_node_y - start_shift],
                            [x, end_node_y - end_shift],
                            [end_node_x, end_node_y]])
            if x == start_node_x:
                line = line[1:]
                line[0, 1] = start_node_y
            if x == end_node_x:
                line = line[:-1]
                line[-1, 1] = end_node_y
            lines.append(line)

            if sorted((n1, n2)) in duplicates:
                line = line.copy()
                n1x, n1y = node_xs_ys[n1]
                n2x, n2y = node_xs_ys[n2]

                lx, hx = sorted([n1x, n2x])
                ly, hy = sorted([n1y, n2y])
                if len(line) == 4:
                    join_1 = n.array([line[2, 0], line[2, 1], 0]) - n.array([line[0, 0], line[0, 1], 0])
                    normal_1 = n.cross(join_1, [0, 0, 1])[:2]
                    normal_1 /= n.linalg.norm(normal_1)

                    join_2 = n.array([line[3, 0], line[3, 1], 0]) - n.array([line[1, 0], line[1, 1], 0])
                    normal_2 = n.cross(join_2, [0, 0, 1])[:2]
                    normal_2 /= n.linalg.norm(normal_2)

                    extra_x_shifts.append(line[1][0] + 0.005 * normal_1[0])

                    line[1] += 0.01*normal_1
                    line[2] += 0.01*normal_2

                elif len(line) == 3:
                    join_1 = n.array([line[2, 0], line[2, 1], 0]) - n.array([line[0, 0], line[0, 1], 0])
                    normal_1 = n.cross(join_1, [0, 0, 1])[:2]
                    normal_1 /= n.linalg.norm(normal_1)

                    extra_x_shifts.append(line[1][0] + 0.005 * normal_1[0])

                    line[1] += 0.01*normal_1
                elif len(line) == 2:
                    join_1 = n.array([line[1, 0], line[1, 1], 0]) - n.array([line[0, 0], line[0, 1], 0])
                    normal_1 = n.cross(join_1, [0, 0, 1])[:2]
                    normal_1 /= n.linalg.norm(normal_1)

                    line = n.vstack([line[0], line[0] + 0.5*(line[1] - line[0]) + 0.01*normal_1, line[1]])

                    extra_x_shifts.append(line[1][0] - 0.005 * normal_1[0])
                
                    
                lines.append(line)

        extra_x_shifts = sorted(extra_x_shifts)[::-1]

        return g, lines, node_labels, nodes_by_height, (leftmost_x, rightmost_x), first_edge, heights, extra_x_shifts

    def draw_planar_graph(self):
        import matplotlib.pyplot as plt
        from matplotlib.patches import Circle
        from matplotlib.collections import PatchCollection

        g, lines, node_labels, nodes_by_height, xlims, first_edge, heights, extra_x_shifts = self._construct_planar_graph()
        leftmost_x, rightmost_x = xlims

        patches = []
        for node, data in g.nodes(data=True):
            y = data['pos']
            xb = data['start']
            xe = data['end']
            x = int((xe + xb) / 2.)
            patches.append(Circle((x, y), 0.25))
        
        plt.ion()

        fig, ax = plt.subplots()
        p = PatchCollection(patches, facecolor='none')
        ax.add_collection(p)
        # plt.axis('equal')

        for node, (x, y) in node_labels.items():
            plt.text(x, y, node,
                     horizontalalignment='center',
                     verticalalignment='center')

        for line in lines:
            ax.plot(line[:, 0], line[:, 1], linewidth=1)
        
        ax.set_xlim(leftmost_x - 1, rightmost_x + 1)
        fig.show()
        
    def space_curve(self, **kwargs):
        from pyknotid.spacecurves import Knot
        self.simplify()

        if len(self) == 0:
            thetas = n.linspace(0, 2*n.pi, 10)
            xs = n.sin(thetas) * 3
            ys = n.cos(thetas) * 3
            zs = n.zeros(10)
            return Knot(n.vstack((xs, ys, zs)).T)
        
        # self.draw_planar_graph()
        g, lines, node_labels, nodes_by_height, xlims, first_edge, heights, extra_x_shifts = self._construct_planar_graph()
        leftmost_x, rightmost_x = xlims

        cg = CrossingGraph()
        for line in lines:
            start_node = nodes_by_height[n.int(n.round(line[0, 1]))]
            end_node = nodes_by_height[n.int(n.round(line[-1, 1]))]
            cl = CrossingLine(start_node, end_node, line)
            cg[start_node].append(cl)
            cg[end_node].append(cl.reversed())

        cg.assert_four_valency()

        cg.align_nodes()
        first_node = 0
        next_node = 1

        points = cg.retrieve_space_curve(
            first_edge[0], first_edge[1], first_edge[2], heights)

        for shift in extra_x_shifts:
            for i in range(len(points)):
                cur_x = points[i, 0]
                if cur_x > shift:
                    points[i, 0] += 1.

        ys = sorted(set(points[:, 1].tolist()))
        for i, y in enumerate(ys):
            points[:, 1][points[:, 1] > y + i + 0.0000001] += 1.

        points[:, 2] *= 1.5

        k = Knot(points*5, verbose=self.verbose)
        k.zero_centroid()
        k.rotate((0.05, 0.03, 0.02))
        return k


class CrossingLine(object):
    def __init__(self, start, end, points):
        self.start = start
        self.end = end
        self.points = points

    def reversed(self):
        return CrossingLine(self.end, self.start, self.points[::-1])

    def __str__(self):
        return '<CrossingLine joining {} with {}>'.format(
            self.start, self.end)

    def __repr__(self):
        return '<CrossingLine joining {} with {}>'.format(
            self.start, self.end)

class CrossingGraph(defaultdict):
    def __init__(self):
        super(CrossingGraph, self).__init__(list)

    def number_of_crossings(self):
        return int(len(self) / 3)
        
    def assert_four_valency(self):
        return True
        for key, value in self.items():
            if len(value) != 4:
                raise ValueError('CrossingGraph is not 4-valent')

    def align_nodes(self):
        '''Orders the lines of each node to be in order, clockwise, depending
        on their incoming angle.
        '''
        for key, value in self.items():
            self[key] = sorted(
                value, key=lambda l: n.arctan2(l.points[1, 1] - l.points[0, 1],
                                               l.points[1, 0] - l.points[0, 0]))

    def retrieve_space_curve(self, first, next, initial_arc_number, heights):
        first_node_lines = self[first]
        for line in first_node_lines:
            if line.end == next:
                break
        else:
            raise ValueError('Node {} is not connected to node {}'.format(first, next))

        possible_start_lines = [l for l in first_node_lines if (l.end == next)]
        if len(possible_start_lines) not in (1, 2):
            raise ValueError('Invalid number of start lines: {}'.format(
                len(possible_start_lines)))

        if len(possible_start_lines) == 1:
            return self._retrieve_space_curve(possible_start_lines[0],
                                              initial_arc_number,
                                              heights)
        else:

            print('possible starts:')
            for line in possible_start_lines:
                print(line)
            
            try:
                return self._retrieve_space_curve(possible_start_lines[0],
                                                  initial_arc_number,
                                                  heights)
            except KeyError as err:
                return self._retrieve_space_curve(possible_start_lines[1],
                                                  initial_arc_number,
                                                  heights)

    def _retrieve_space_curve(self, line, initial_arc_number, heights):
        current_line = line
        segments = []
        h = 1.
        arc_number = initial_arc_number
        for _ in range(4*self.number_of_crossings()):
        # for _ in range(len(self)*2):
            current_points = current_line.points.copy()
            ps = n.zeros((len(current_points), 3))
            ps[:, :-1] = current_points

            height = heights[(current_line.start, current_line.end, arc_number)]
            ps[0, -1] = height[0] # height
            # ps[0, -1] = h; h *= -1.

            segments.append(ps[:-1])

            next_lines = self[current_line.end]
            incoming_angle = n.arctan2(current_points[-2, 1] - current_points[-1, 1],
                                       current_points[-2, 0] - current_points[-1, 0])

            other_incoming_angles = [
                n.arctan2(l.points[1, 1] - l.points[0, 1],
                          l.points[1, 0] - l.points[0, 0]) for l in next_lines]

            angle_distances = [
                angle_distance(angle, incoming_angle)
                for angle in other_incoming_angles]

            incoming_index = n.argmin(angle_distances)
            if len(angle_distances) == 4:
                outgoing_index = (incoming_index + 2) % 4 
            elif len(angle_distances) == 2:
                outgoing_index = (incoming_index + 1) % 2
            else:
                raise ValueError('Encountered node with neither 2 nor 4 arcs, should be impossible')
            current_line = next_lines[outgoing_index]

            # This is the old arc number calculation before adding intermediates
            # arc_number = (arc_number % (len(self) * 2)) + 1

            if len(angle_distances) == 4:
                arc_number = (arc_number % (2*self.number_of_crossings())) + 1

        return n.vstack(segments)

def angle_distance(a1, a2):
    dist = n.abs(a2 - a1)
    if dist > n.pi:
        dist = 2*n.pi - dist
    return dist




'''
Link
====

.. image:: torus_hopf_link.png
   :align: center
   :scale: 50%
   :alt: A hopf link visualised by pyknotid

Class for dealing with multiple curves as a link. :class:`Link`
provides methods for topological manipulation and calculations on
multiple curves.


API documentation
~~~~~~~~~~~~~~~~~
'''

import numpy as n

try:
    from pyknotid.spacecurves import chelpers
except:
    from pyknotid.spacecurves import helpers as chelpers
from pyknotid.spacecurves import helpers
from pyknotid.spacecurves.knot import Knot
from pyknotid.visualise import plot_projection
from pyknotid.utils import (vprint, get_rotation_matrix,
                           ensure_shape_tuple)


class Link(object):
    '''
    Class for holding the vertices of multiple lines, providing helper
    methods for convenient manipulation and analysis.

    The data is stored
    internally as multiple :class:`Knot`s.

    Parameters
    ----------
    lines : list of nx3 array-like or Knots
        List with the points of each line.
    verbose : bool
        Whether to print information during processing. Defaults
        to True.
    '''

    def __init__(self, lines, verbose=True):
        self._lines = []
        self.verbose = verbose

        if isinstance(lines, Link):
            self.lines = [Knot(line) for line in lines.lines]
        else:
            lines = [Knot(line) for line in lines]
            self.lines = lines

        self._crossings = None
        self._gauss_code = {}

        self._recent_octree = None

    @property
    def lines(self):
        return self._lines

    @lines.setter
    def lines(self, lines):
        self._lines = lines
        self._reset_cache()

    def _reset_cache(self):
        self._crossings = None
        self._gauss_code = {}

    @classmethod
    def from_periodic_lines(cls, lines, shape, perturb=True):
        '''Returns a :class:`Link` instance in which the lines have
        been unwrapped through the periodic boundaries.

        Parameters
        ----------
        line : list
            A list of the Nx3 vectors of points in the lines
        shape : array-like
            The x, y, z distances of the periodic boundary
        perturb : bool
            If True, translates and rotates the knot to avoid any lattice
            problems.
        '''
        shape = ensure_shape_tuple(shape)
        lines = [Knot.from_periodic_line(line, shape, perturb=False)
                 for line in lines]
        link = cls(lines)
        if perturb:
            link.translate(n.array([0.00123, 0.00231, 0.00321]))
            link.rotate()
        return link

    def raw_crossings(self, mode='use_max_jump', only_with_other_lines=True,
                      include_closures=True,
                      recalculate=False,
                      try_cython=True):
        '''Returns the crossings in the diagram of the projection of the
        space curve into its z=0 plane.

        The crossings will be calculated the first time this function
        is called, then cached until an operation that would change
        the list (e.g. rotation, or changing ``self.points``).

        Multiple modes are available (see parameters) - you should be
        aware of this because different modes may be vastly slower or
        faster depending on the type of line.

        Parameters
        ----------
        mode : str, optional
            One of ``'count_every_jump'`` and ``'use_max_jump'``. In the former
            case,
            walking along the line uses information about the length of
            every step. In the latter, it guesses that all steps have the
            same length as the maximum step length. The optimal choice
            depends on the data, but is usually ``'use_max_jump'``, which
            is the default.
        only_with_other_lines : bool
            If True, ignores self-crossings (i.e. the knot type of the loops)
            and returns only a list of crossings between the loops. Defaults
            to True
        include_closures : bool, optional
            Whether to include crossings with the
            lines joining their start and end points. Defaults to True.
        recalculate : bool, optional
            Whether to force a recalculation of the crossing positions.
            Defaults to False.
        try_cython : bool, optional
            Whether to try to use a cython implementation of crossing
            finding. This will make no difference if the cython could not
            be loaded, in which case python is already
            used automatically. Defaults to True.

        Returns
        -------
        array-like
            The raw array of floats representing crossings, of the
            form [[line_index, other_index, +-1, +-1], ...], where the
            line_index and other_index are in arclength parameterised
            by integers for each vertex and linearly interpolated,
            and the +-1 represent over/under and clockwise/anticlockwise
            respectively.
        '''

        if (not recalculate and self._crossings is not None and
            self._crossings[0] == only_with_other_lines):
            return self._crossings[1]

        if try_cython:
            helpers_module = chelpers
        else:
            helpers_module = helpers

        lines = self.lines

        # Get length of each line
        line_lengths = [0.]
        line_lengths.extend([line.arclength() for line in lines])
        cumulative_lengths = n.cumsum(line_lengths)[:-1]

        if only_with_other_lines:
            crossings = [[] for _ in lines]
        else:
            self._vprint('Calculating self-crossings for all {} '
                         'component lines'.format(len(lines)))
            crossings = [k.raw_crossings(
                mode=mode, include_closure=include_closures,
                recalculate=recalculate, try_cython=try_cython) for k in lines]
            for index, cum_length in enumerate(cumulative_lengths):
                if len(crossings[index]):
                    crossings[index][:, :2] += cum_length
                crossings[index] = crossings[index].tolist()

        jump_mode = {'count_every_jump': 1, 'use_max_jump': 2,
                     'naive': 3}[mode]

        segment_lengths = [
            n.roll(line.points[:, :2], -1, axis=0) - line.points[:, :2] for
            line in lines]
        segment_lengths = [
            n.sqrt(n.sum(lengths * lengths, axis=1)) for
            lengths in segment_lengths]

        if include_closures:
            max_segment_length = n.max(n.hstack(segment_lengths))
        else:
            max_segment_length = n.max(n.hstack([
                lengths[:-1] for lengths in segment_lengths]))

        for line_index, line in enumerate(lines):
            for other_index, other_line in enumerate(lines[line_index+1:]):
                self._vprint(
                    '\rComparing line {} with {}'.format(line_index,
                                                         other_index))

                other_index += line_index + 1

                points = line.points
                comparison_points = other_line.points
                if include_closures:
                    comparison_points = n.vstack((comparison_points,
                                                  comparison_points[:1]))

                other_seg_lengths = segment_lengths[other_index]
                # other_seg_lengths is already corrected to include
                # closures if necessary

                first_line_range = range(len(points))
                if not include_closures:
                    first_line_range = first_line_range[:-1]

                for i in first_line_range:
                    if i % 1000 == 0:
                        self._vprint(
                            '\ri = {} / {}'.format(
                                i, len(first_line_range)), False)
                    v0 = points[i]
                    dv = points[(i + 1) % len(points)] - v0

                    vnum = i
                    compnum = 0  # start at beginning of other line

                    new_crossings = helpers_module.find_crossings(
                        v0, dv, comparison_points, other_seg_lengths,
                        vnum, compnum,
                        max_segment_length,
                        jump_mode)

                    if not len(new_crossings):
                        continue
                    first_crossings = n.array(new_crossings[::2])
                    first_crossings[:, 0] += cumulative_lengths[line_index]
                    first_crossings[:, 1] += cumulative_lengths[other_index]
                    sec_crossings = n.array(new_crossings[1::2])
                    sec_crossings[:, 0] += cumulative_lengths[other_index]
                    sec_crossings[:, 1] += cumulative_lengths[line_index]

                    crossings[line_index].extend(first_crossings.tolist())
                    crossings[other_index].extend(sec_crossings.tolist())

        self._vprint('\n{} crossings found\n'.format(
            [len(cs) / 2 for cs in crossings]))
        [cs.sort(key=lambda s: s[0]) for cs in crossings]
        crossings = [n.array(cs) for cs in crossings]
        self._crossings = (only_with_other_lines, crossings)

        return crossings

    def translate(self, vector, lines=None):
        '''Translate all points in some or all lines of self.

        Parameters
        ----------
        vector : array-like
            The x, y, z translation distances
        lines : list or int
            The list of line indices to which the translation should
            be applied. Defaults to None, which applies the translation
            to all the lines of self. If an integer is supplied, only
            the line with this index is translated.
        '''
        if lines is None:
            lines = range(len(self.lines))
        elif isinstance(lines, int):
            lines = [lines]
        for index, line in enumerate(self.lines):
            if index in lines:
                line.translate(vector)

    def rotate(self, angles=None):
        '''
        Rotates all the points of each line of self by the given angle
        in each axis.

        Parameters
        ----------
        angles : array-like
            Rotation angles about x, y and z axes. If None, random angles
            are used. Defaults to None.
        '''
        if angles is None:
            angles = n.random.random(3)
        for line in self.lines:
            line.rotate(angles)
        self._reset_cache()

    def plot(self, mode='vispy', clf=True, colours=None, **kwargs):
        '''
        Plots all the lines. See :func:`pyknotid.visualise.plot_line` for
        full documentation.
        '''
        lines = self.lines

        if 'closed' not in kwargs:
            kwargs['closed'] = True

        if mode =='vispy':
            from pyknotid.visualise import plot_cell_vispy
            points = [k.points for k in lines]
            plot_cell_vispy([[p] for p in points], boundary=None,
                            **kwargs)
            return

        if colours is not None:
            kwargs.update({'color': colours[0]})
        lines[0].plot(mode=mode, clf=clf, **kwargs)
        for i, line in enumerate(lines[1:]):
            if colours is not None:
                kwargs.update({'color': colours[i+1]})
            line.plot(mode=mode, clf=False, zero_centroid=False, **kwargs)

    def plot_projection(self, with_crossings=True, mark_start=False,
                        include_self_crossings=False):
        all_points = [line.points for line in self.lines]
        lengths = [k.arclength() for k in self.lines]
        cum_lengths = n.hstack([[0], n.cumsum(lengths)])

        crossings = None
        plot_crossings = []
        if with_crossings:
            all_crossings = self.raw_crossings(
                only_with_other_lines=(not include_self_crossings))
            plot_crossings = []
            for index, crossings in enumerate(all_crossings):
                remove_length = cum_lengths[index]
                points = all_points[index]
                for crossing in crossings:
                    x, y, over, orientation = crossing

                    # Work out which line the crossing is on
                    # next_x_start = n.argmax(cum_lengths > x)
                    # x_line = next_x_start - 1
                    # x -= cum_lengths[x_line]
                    x -= remove_length

                    xint = int(x)
                    r = points[xint]
                    dr = points[(xint+1) % len(points)] - r
                    plot_crossings.append(r + (x-xint) * dr)
        fig, ax = plot_projection(all_points[0],
                                  crossings=n.array(plot_crossings),
                                  mark_start=mark_start)
        for line in all_points[1:]:
            plot_projection(line,
                            crossings=n.array(plot_crossings),
                            mark_start=mark_start,
                            fig_ax=(fig, ax))

        ax.autoscale(tight=False)
        return fig, ax

    def octree_simplify(self, runs=1, plot=False, rotate=True,
                        obey_knotting=False, **kwargs):
        '''
        Simplifies the curves via the octree reduction of
        :module:`pyknotid.simplify.octree`.

        Parameters
        ----------
        runs : int
            The number of times to run the octree simplification.
            Defaults to 1.
        plot : bool
            Whether to plot the curve after each run. Defaults to False.
        rotate : bool
            Whether to rotate the space curve before each run. Defaults
            to True as this can make things much faster.
        obey_knotting : bool
            Whether to not let the line pass through itself. Defaults to
            False - knotting of individual components will be ignored!
            This is *much* faster than the alternative.

        kwargs are passed to the :class:`pyknotid.simplify.octree.OctreeCell`
        constructor.
        '''
        from ..simplify.octree import OctreeCell, remove_nearby_points
        for line in self.lines:
            line.points = remove_nearby_points(line.points)
        for i in range(runs):
            if n.sum([len(knot.points) for knot in self.lines]) > 30:
                vprint('\rRun {} of {}, {} points remain'.format(
                    i, runs, len(self)), False, self.verbose)

            if rotate:
                rot_mat = get_rotation_matrix(n.random.random(3))
                for line in self.lines:
                    line._apply_matrix(rot_mat)

            oc = OctreeCell.from_lines([line.points for line in self.lines],
                                       **kwargs)
            oc.simplify(obey_knotting)
            self._recent_octree = oc
            self.lines = [Knot(line) for line in oc.get_lines()]
            for line in self.lines:
                line.points = remove_nearby_points(line.points)

            if rotate:
                for line in self.lines:
                    line._apply_matrix(rot_mat.T)

            if plot:
                self.plot()

        vprint('\nReduced to {} points'.format(len(self)))

    def __len__(self):
        return n.sum(map(len, self.lines))

    def arclength(self, include_closures=True):
        '''
        Returns the sum of arclengths of the lines.

        Parameters
        ----------
        include_closures : bool
            Whether to include the distance between the final and
            first points of each line. Defaults to True.
        '''
        return n.sum(k.arclength(include_closures) for k in self.lines)

    def _vprint(self, s, newline=True):
        '''Prints s, with optional newline. Intended for internal use
        in displaying progress.'''
        if self.verbose:
            vprint(s, newline)
        
    def gauss_code(self, **kwargs):
        '''
        Returns a :class:`~pyknotid.representations.gausscode.GaussCode`
        instance representing the crossings of the knot.

        The GaussCode instance is cached internally. If you want to
        recalculate it (e.g. to get an unsimplified version if you
        have simplified it), you should pass `recalculate=True`.

        This method passes kwargs directly to :meth:`raw_crossings`,
        see the documentation of that function for all options.
        '''

        from ..representations.gausscode import GaussCode
        include_self_linking = not kwargs.get('only_with_other_lines', True)
        if ('recalculate' not in kwargs and
            include_self_linking in self._gauss_code):
            return self._gauss_code[include_self_linking]
        crossings = self.raw_crossings(**kwargs)
        gc = GaussCode(crossings)
        self._gauss_code[include_self_linking] = gc
        return gc

    def linking_number(self, **kwargs):
        '''
        Returns the linking number of the lines in the Link, the
        sum of signed crossings between them, ignoring crossings of
        a line with itself.
        '''
        crossings = self.raw_crossings(only_with_other_lines=True,
                                       **kwargs)
        number = 0
        for line in crossings:
            if len(line):
                number += n.sum(line[:, 3])
        return int(n.abs(number / 2))

    def smooth(self, *args, **kwargs):
        '''
        Smooths each of the x, y and z components of each of self.lines
        by convolving with a window of the given type and size.

        kwargs are passed straight to
        :meth:`pyknotid.spacecurves.spacecurve.SpaceCurve.smooth`.
        '''
        for line in self.lines:
            if len(line) > 5:
                line.smooth(*args, **kwargs)

    def multivariate_alexander(self, variables=-1., **kwargs):

        from ..invariants import multivariate_alexander_numpy
        gc = self.gauss_code(**kwargs)
        gc.simplify(verbose=self.verbose)
        return multivariate_alexander(gc, variables)

        



'''
Building the database
=====================

.. warning:: This module is only used for building the database. It is
             generally recommended to download the prebuilt database.

Contains functions for improving an existing database by filling in
invariants. Original motivation was/is to get Jones polynomials from
the HOMFLY polynomials provided in the high-crossing data, but now
includes routines to get fiberedness etc from KnotInfo, and to
calculate new invariants based on the DT notation.

'''

from __future__ import print_function

from pyknotid.catalogue.database import Knot, db
from pyknotid.catalogue.converters import homfly_to_jones, db2py_homfly, py2db_jones
from pyknotid.catalogue.build import get_rdf_object

import csv
import json
import numpy as n


def planar_writhe_from_dt_code(max_crossings=12):

    from pyknotid.catalogue import from_invariants
    from pyknotid.representations import dtnotation as dt

    knots = Knot.select().where(
        Knot.min_crossings <= max_crossings).where(
            Knot.planar_writhe >> None).where(
                Knot.min_crossings > 0)

    output_knots = []
    i = 0
    for knot in knots:
        if i % 100 == 0:
            print(i, len(output_knots))

        i += 1

        assert knot.planar_writhe is None
        assert knot.min_crossings <= max_crossings

        d = dt.DTNotation(knot.dt_code)
        r = d.representation()
        if r is None:
            print('{} caused trouble, skipping'.format(knot))
            continue

        writhe = r.writhe()
        if writhe < 0:
            writhe *= -1
        knot.planar_writhe = writhe
        output_knots.append(knot)

        if i % 10000 == 0:
            print('Saving changes')
            with db.transaction():
                for knot in output_knots:
                    knot.save()
                output_knots = []

        print(knot, writhe)

    print('Saving changes')
    with db.transaction():
        for knot in output_knots:
            knot.save()
        output_knots = []
 
def jones_from_homfly():
    '''Takes any knots with a homfly polynomial but lacking jones, and
    computes the latter from the former.'''

    knots = Knot.select().where(Knot.jones >> None)

    output_knots = []
    i = 0
    for knot in knots:
        if i % 100 == 0:
            print(i, len(output_knots))
        i += 1
        if knot.jones is None and knot.homfly is not None:
            homfly = knot.homfly
            if 'Fail' not in homfly:
                homfly = db2py_homfly(homfly)
                jones = homfly_to_jones(homfly)
                jones = py2db_jones(jones)
                knot.jones = jones
                output_knots.append(knot)

        if i % 10000 == 0:
            print('Saving changes')
            with db.transaction():
                for knot in output_knots:
                    knot.save()
            output_knots = []

    print('Saving changes')
    with db.transaction():
        for knot in output_knots:
            knot.save()

def alexander_imags_from_alexander(min_crossings=None):
    '''Takes any knots with an Alexander polynomial but lacking its values
    at exp(2 pi I / 3) and exp(2 pi I / 4). It could also do the
    determinant (some of the determinants in the knot atlas data are
    wrong), but doesn't for now.
    '''
    knots = Knot.select().where(~(Knot.alexander >> None))
    if min_crossings is not None:
        knots = knots.where(Knot.min_crossings == min_crossings)

    output_knots = []
    for i, knot in enumerate(knots):
        if i % 100 == 0:
            print(i, len(output_knots))
        if knot.alexander is not None:
            array = json.loads(knot.alexander)

            exp_2s = [(-1.)**i for i in range(len(array))]
            alexander_imag_2 = int(n.round(n.abs(n.sum([
                coefficient * exponential for coefficient, exponential
                in zip(array, exp_2s)]))))
            
            exp_3s = [n.exp(i*2*n.pi*1j/3.) for i in range(len(array))]
            alexander_imag_3 = int(n.round(n.abs(n.sum([
                coefficient * exponential for coefficient, exponential
                in zip(array, exp_3s)]))))

            exp_4s = [(1.j)**i for i in range(len(array))]
            alexander_imag_4 = int(n.round(n.abs(n.sum([
                coefficient * exponential for coefficient, exponential
                in zip(array, exp_4s)]))))

            knot.determinant = alexander_imag_2
            knot.alexander_imag_3 = alexander_imag_3
            knot.alexander_imag_4 = alexander_imag_4
            output_knots.append(knot)

        if i % 10000 == 0:
            print('Saving changes')
            with db.transaction():
                for knot in output_knots:
                    knot.save()
            output_knots = []
    print('Saving changes')
    with db.transaction():
        for knot in output_knots:
            knot.save()
    output_knots = []
        
            
def add_fiberedness():
    '''Adds fiberedness information to any knots where the information is
    available in knotinfo's database.'''
    data = []
    with open('/home/asandy/knotcatalogue/raw_data/knotinfo_data_complete.csv','r') as fileh:
        reader = csv.reader(fileh, delimiter=',', quotechar='"')
        for row in reader:
            entries = []
            for column in row:
                entries.append(column)
            data.append(entries)

    output_knots = []
    for row in data[2:]:
        name = row[0]
        monodromy = row[158]
        fibered = False
        if monodromy != 'Not Fibered':
            fibered = True

        crossings, number = name.split('_')
        number = int(number)
        if int(crossings[:2]) >= 11:
            realname = 'K{}{}'.format(crossings, number)
        else:
            crossings = int(crossings)
            realname = '{}_{}'.format(crossings, number)
        knots = list(Knot.select().where(Knot.identifier==realname))
        if len(knots) != 1:
            print('Failed to find {} in db'.format(realname))
        else:
            knot = knots[0]
            knot.fibered = fibered
            output_knots.append(knot)

    print('Attempting to save')
    with db.transaction():
        for knot in output_knots:
            knot.save()

def add_vassilievs_and_symmetry_from_rdf(filen):
    '''Adds Vassiliev 2 and 3 from RDF files. This is only necessary for
    pre-existing databases, these invariants are now added by
    build.py.
    '''

    from rdflib import URIRef, Graph
    rdfv2 = URIRef('invariant:V_2')
    rdfv3 = URIRef('invariant:V_3')
    rdfsymmetry = URIRef('invariant:Symmetry_Type')

    g = Graph()
    g.parse(filen, format='nt')

    subjects = list(set(g.subjects(None, None)))

    output_knots = []
    for subject in subjects:
        identifier = str(subject.toPython().split(':')[1])

        knots = Knot.select().where(Knot.identifier == identifier)
        first = knots.first()
        if first is not None:
            v2 = get_rdf_object(g, subject, rdfv2)
            if v2 is not None:
                v2 = int(v2)
            v3 = get_rdf_object(g, subject, rdfv3)
            if v3 is not None:
                v3 = int(v3)
            symmetry = get_rdf_object(g, subject, rdfsymmetry)
            if symmetry is not None:
                symmetry = symmetry.lower()
            first.vassiliev_2 = v2
            first.vassiliev_3 = v3
            first.symmetry = symmetry
            output_knots.append(first)
            print('Added {}; {}, {}, {}'.format(identifier, v2, v3, symmetry))
        else:
            print('Failed to find {} in db'.format(identifier))
                
    print('Attempting to save')
    with db.transaction():
        for knot in output_knots:
            knot.save()
        
def add_two_bridgeness():
    '''Adds two-bridge information to any knots where the information is
    available in knotinfo's database.
    '''
    data = []
    with open('/home/asandy/knotcatalogue/raw_data/knotinfo_data_complete.csv','r') as fileh:
        reader = csv.reader(fileh, delimiter=',', quotechar='"')
        for row in reader:
            entries = []
            for column in row:
                entries.append(column)
            data.append(entries)

    output_knots = []
    for row in data[2:]:
        name = row[0]
        twobridgenotation = row[24]
        if twobridgenotation == '':
            twobridgenotation = None

        crossings, number = name.split('_')
        number = int(number)
        if int(crossings[:2]) >= 11:
            realname = 'K{}{}'.format(crossings, number)
        else:
            crossings = int(crossings)
            realname = '{}_{}'.format(crossings, number)
        knots = list(Knot.select().where(Knot.identifier==realname))
        if len(knots) != 1:
            print('Failed to find {} in db'.format(realname))
        else:
            knot = knots[0]
            knot.two_bridge = twobridgenotation
            output_knots.append(knot)

    print('Attempting to save')
    with db.transaction():
        for knot in output_knots:
            knot.save()
        

    
def check_determinants(**parameters):
    from pyknotid.catalogue import from_invariants

    knots = from_invariants(**parameters)

    fails = []

    for knot in knots:
        k = knot.space_curve(verbose=False)
        det = k.determinant()

        if knot.determinant != det:
            print('{}: FAIL, calculated {} vs {} in db'.format(
                knot, det, knot.determinant))
            fails.append(knot)
        else:
            print('{}: SUCCESS, {}'.format(knot, det))

    return fails


def vassiliev_invariants_from_space_curve(max_crossings=12):   

    knots = Knot.select().where(Knot.vassiliev_3 >> None).where(
        Knot.min_crossings <= max_crossings)
    print('Calculating Vassiliev 3 for {} knots'.format(knots.count()))

    output_knots = []
    num_knots = knots.count()
    for i, knot in enumerate(knots):
        if i % 1 == 0:
            print(i, '/', num_knots)

        assert knot.vassiliev_3 is None

        ks = knot.space_curve(verbose=False)
        v3 = ks.vassiliev_degree_3()

        knot.vassiliev_3 = v3
        output_knots.append(knot)

        if knot.vassiliev_2 is None:
            v2 = ks.vassiliev_degree_2()
            knot.vassiliev_2 = v2
        else:
            v2 = knot.vassiliev_2

        if i % 10000 == 0:
            print('saving changes')
            with db.transaction():
                for knot in output_knots:
                    knot.save()
                output_knots = []

        print(knot, v2, v3)

    print('saving changes')
    with db.transaction():
        for knot in output_knots:
            knot.save()
        output_knots = []


def vassiliev_2_invariants_from_space_curve(max_crossings=16):   

    knots = Knot.select().where(Knot.vassiliev_2 >> None).where(
        Knot.min_crossings <= max_crossings)
    print('Calculating Vassiliev 2 for {} knots'.format(knots.count()))

    output_knots = []
    num_knots = knots.count()
    for i, knot in enumerate(knots):
        if i % 1 == 0:
            print(i, '/', num_knots)

        assert knot.vassiliev_2 is None

        ks = knot.space_curve(verbose=False)
        v2 = ks.vassiliev_degree_2()

        knot.vassiliev_2 = v2
        output_knots.append(knot)

        print(knot, v2)

    print('saving changes')
    with db.transaction():
        for knot in output_knots:
            knot.save()
        output_knots = []

        
def alexander_at_roots_from_dt_code():
    knots = Knot.select().where(Knot.alexander_imag_3 >> None)

    output_knots = []
    
    num_knots = knots.count()
    for i, knot in enumerate(knots):
        if i % 1 == 0:
            print(i, '/', num_knots)

        ks = knot.space_curve(verbose=False)

        roots = ks.alexander_at_root((2, 3, 4))

        knot.determinant = roots[0]
        knot.alexander_imag_3 = roots[1]
        knot.alexander_imag_4 = roots[2]

        output_knots.append(knot)

        print(knot, roots)

    print('saving changes')
    with db.transaction():
        for knot in output_knots:
            knot.save()

        



'''
Building a knot database
========================

Functions for building a knot database from raw data
files. Intended for use with the RDF format data from the Knot
Atlas.'''

import sympy as sym
import rdflib
from rdflib import URIRef, Graph

import database as db
from database import Knot
import converters

rdfalex = URIRef('invariant:Alexander_Polynomial')
rdfjones = URIRef('invariant:Jones_Polynomial')
rdfhomfly = URIRef('invariant:HOMFLYPT_Polynomial')

rdfcross = URIRef('invariant:Crossings')
rdfdet = URIRef('invariant:Determinant')
rdfsig = URIRef('invariant:Signature')

rdfunknot = URIRef('invariant:Unknotting_Number')
rdfhyperbolic = URIRef('invariant:HyperbolicVolume')

rdfgc = URIRef('invariant:Gauss_Code')
rdfconway = URIRef('invariant:ConwayNotation')
rdfpd = URIRef('invariant:PD_Presentation')
rdfdtcode = URIRef('invariant:DT_Code')

rdfv2 = URIRef('invariant:V_2')
rdfv3 = URIRef('invariant:V_3')

rdfsymmetry = URIRef('invariant:Symmetry_Type')

# RDFLib arrangement is subject/predicate/object

db.db.get_conn()

# Try to create a Knot table in the database, just in case one doesn't
# exist. This might be a silly way to deal with things, but it'll do
# for now.
if not db.Knot.table_exists():
    db.Knot.create_table()


def add_knots_from(filen):
    '''Read the RDF file at filen, and add all its knots to the database
    specified in database.py.

    The filen *must* be formatted in rdf nt format. This is the case
    for knot atlas take home database files.

    '''
    g = Graph()
    g.parse(filen, format='nt')

    subjects = list(set(g.subjects(None, None)))
    # Using the set removes duplicates

    knots = []
    total = len(subjects)
    i = 0
    for subject in subjects:
        i += 1
        if i % 10 == 0:
            sys.stdout.write('\r{}: knot {} of {}'.format(filen, i, total))
            sys.stdout.flush()

        # Each subject is a knot
        identifier = str(subject.toPython().split(':')[1])

        alexander = get_rdf_object(g, subject, rdfalex)
        if alexander is not None:
            alexander = converters.rdf_poly_to_sympy(alexander, 't')
            alexander = converters.alexander_to_json(alexander)

        jones = get_rdf_object(g, subject, rdfjones)
        if jones is not None:
            jones = converters.rdf_poly_to_sympy(jones, 'q')
            jones = converters.jones_to_json(jones)

        homfly = get_rdf_object(g, subject, rdfhomfly)
        #print 'homfly is', homfly
        if homfly is not None and 'Failed' not in homfly:
            homfly = homfly.replace('\\text{QuantumGroups$\\grave{ }$', '')
            homfly = homfly.replace('a}', 'a')
            homfly = converters.rdf_poly_to_sympy(homfly, ['z','a'])
            homfly = converters.homfly_to_json(homfly)

        crossings = get_rdf_object(g, subject, rdfcross)
        if crossings is not None:
            crossings = int(crossings[0])
        else:
            crossings = int(identifier.split('_')[0][1:3])

        determinant = get_rdf_object(g, subject, rdfdet)
        if determinant is not None:
            determinant = int(determinant)

        signature = get_rdf_object(g, subject, rdfsig)
        if signature is not None:
            signature = int(signature)

        unknot_number = get_rdf_object(g, subject, rdfunknot)
        if unknot_number is not None:
            if 'math' in unknot_number:
                unknot_number = None
            else:
                unknot_number = int(unknot_number)

        hyp_vol = get_rdf_object(g, subject, rdfhyperbolic)
        if hyp_vol is not None:
            hyp_vol = str(hyp_vol)

        gauss_code = get_rdf_object(g, subject, rdfgc)
        if gauss_code is not None:
            gauss_code = str(gauss_code)

        conway = get_rdf_object(g, subject, rdfconway)
        if conway is not None:
            conway = str(conway)
            conway = conway.replace('<nowiki>', '')
            conway = conway.replace('</nowiki>', '')

        planar_diagram = get_rdf_object(g, subject, rdfpd)
        if planar_diagram is not None:
            planar_diagram = str(planar_diagram)
            planar_diagram = planar_diagram.replace('<sub>', '_')
            planar_diagram = planar_diagram.replace('</sub>', '')

        dt_code = get_rdf_object(g, subject, rdfdtcode)
        if dt_code is not None:
            dt_code = str(dt_code)

        v2 = get_rdf_object(g, subject, rdfv2)
        if v2 is not None:
            v2 = int(v2)

        v3 = get_rdf_object(g, subject, rdfv3)
        if v3 is not None:
            v3 = int(v3)

        symmetry = get_rdf_object(g, subject, rdfsymmetry)
        if symmetry is not None:
            symmetry = symmetry.lower()
        

        k = Knot(identifier=identifier,
                 min_crossings=crossings,
                 determinant=determinant,
                 signature=signature,
                 alexander=alexander,
                 jones=jones,
                 homfly=homfly,
                 unknotting_number=unknot_number,
                 hyperbolic_volume=hyp_vol,
                 conway_notation=conway,
                 gauss_code=gauss_code,
                 planar_diagram=planar_diagram,
                 dt_code=dt_code,
                 vassiliev_2=v2,
                 vassiliev_3=v3,
                 symmetry=symmetry,
                 )
        knots.append(k)
    sys.stdout.write('\n')
    sys.stdout.flush()

    print 'Attempting to save in transaction...'
    with db.db.transaction():
        for knot in knots:
            knot.save()
    return knots


def get_rdf_object(graph, subject, predicate):
    '''Takes an rdflib Graph, subject and predicate, and returns the first
    matching object if one exists. If none exist, returns None.'''
    objects = list(graph.objects(subject, predicate))
    if len(objects) == 0:
        return None
    else:
        return objects[0].toPython()

if __name__ == '__main__':
    import sys

    filens = sys.argv[1:]
    for filen in filens:
        print 'Reading in from', filen
        add_knots_from(filen)



'''
Knot
----

Class for dealing with curves as knots. :class:`Knot` provides many
methods for topological manipulation and calculations.

API documentation
~~~~~~~~~~~~~~~~~

'''

from __future__ import division

import numpy as n
import numpy as np

from pyknotid.spacecurves.spacecurve import SpaceCurve

__all__ = ('Knot', )


class Knot(SpaceCurve):
    '''
    Class for holding the vertices of a single line, providing helper
    methods for convenient manipulation and analysis.

    A :class:`Knot` just represents a single space curve, it may be
    topologically trivial!

    This class deliberately combines methods to do many different kinds
    of measurements or manipulations. Some of these are externally
    available through other modules in pyknotid - if so, this is usually
    indicated in the method docstrings.

    Parameters
    ----------
    points : array-like
        The 3d points (vertices) of a piecewise
        linear curve representation
    verbose : bool
        Indicates whether the Knot should print
        information during processing
    add_closure : bool
        If True, adds a final point to the knot near to the start point,
        so that it will appear visually to close when plotted.
    '''

    @SpaceCurve.points.setter
    def points(self, points):
        super(Knot, self.__class__).points.fset(self, points)
        self._cached_isolated = None

    def tangents(self):
        return super(Knot, self).tangents(closed=True)

    def copy(self):
        '''Returns another knot with the same points and verbosity
        as self. Other attributes (e.g. cached crossings) are not
        preserved.'''
        return Knot(self.points.copy(), verbose=self.verbose)

    def plot(self, **kwargs):
        super(Knot, self).plot(closed=True, **kwargs)

    def reconstructed_space_curve(self):
        r = self.representation()
        k = Knot(r.space_curve())
        k.zero_centroid()
        return k

    def alexander_polynomial(self, variable=-1, quadrant='lr',
                             mode='python', **kwargs):
        '''
        Returns the Alexander polynomial at the given point,
        as calculated by :func:`pyknotid.invariants.alexander`.

        See :func:`pyknotid.invariants.alexander` for the meanings
        of the named arguments.
        '''
        from ..invariants import alexander
        gc = self.gauss_code(**kwargs)
        gc.simplify()
        return alexander(gc, variable=variable, quadrant=quadrant,
                         simplify=False, mode=mode)

    def alexander_at_root(self, root, round=True, **kwargs):
        '''
        Returns the Alexander polynomial at the given root of unity,
        i.e. evaluated at exp(2 pi I / root).

        The result returned is the absolute value.

        Parameters
        ----------
        root : int
            The root of unity to use, i.e. evaluating at exp(2 pi I / root).
            If this is iterable, this method returns a list of the results
            at every value of that iterable.
        round : bool
            If True and n in (1, 2, 3, 4), the result will be rounded
            to the nearest integer for convenience, and returned as an
            integer type.
        **kwargs :
            These are passed directly to :meth:`alexander_polynomial`.
        '''
        if hasattr(root, '__contains__'):
            return [self.alexander_at_root(r) for r in root]
        variable = n.exp(2 * n.pi * 1.j / root)
        value = self.alexander_polynomial(variable, **kwargs)
        value = n.abs(value)
        if round and root in (1, 2, 3, 4):
            value = int(n.round(value))
        return value

    def determinant(self):
        '''
        Returns the determinant of the knot. This is the Alexander
        polynomial evaluated at -1.
        '''
        return self.alexander_at_root(2)

    def vassiliev_degree_2(self, simplify=True, **kwargs):
        '''
        Returns the Vassiliev invariant of degree 2 for the Knot.

        Parameters
        ==========
        simplify : bool
            If True, simplifies the Gauss code of self before
            calculating the invariant. Defaults to True, but
            will work fine if you set it to False (and might even
            be faster).
        **kwargs :
            These are passed directly to :meth:`gauss_code`.
        '''
        from ..invariants import vassiliev_degree_2
        gc = self.gauss_code(**kwargs)
        if simplify:
            gc.simplify()
        return vassiliev_degree_2(gc)

    def vassiliev_degree_3(self, simplify=True, try_cython=True, **kwargs):
        '''Returns the Vassiliev invariant of degree 3 for the Knot.

        Parameters
        ==========
        simplify : bool
            If True, simplifies the Gauss code of self before
            calculating the invariant. Defaults to True, but
            will work fine if you set it to False (and might even
            be faster).
        try_cython : bool
            Whether to try and use an optimised cython version of the
            routine (takes about 1/3 of the time for complex
            representations).  Defaults to True, but the python
            fallback will be *slower* than setting it to False if the
            cython function is not available.
        **kwargs :
            These are passed directly to :meth:`gauss_code`.

        '''
        from ..invariants import vassiliev_degree_3
        gc = self.gauss_code(**kwargs)
        if simplify:
            gc.simplify()
        return vassiliev_degree_3(gc, try_cython=try_cython)

    def hyperbolic_volume(self):
        '''
        Returns the hyperbolic volume at the given point, via
        :meth:`pyknotid.representations.PlanarDiagram.as_spherogram`.

        Returns
        -------
        volume : float
            A float representing the volume returned.
        accuracy : int
            The number of digits of precision. This is significant
            digits, e.g. 0.00021 with 1 digit precision = 2E-4.
        solution_type : str
            The solution type of the manifold. Normally one of:
            - 'contains degenerate tetrahedra' => may not be a valid result
            - 'all tetrahedra positively oriented' =>
              really probably hyperbolic
        '''
        from ..invariants import hyperbolic_volume
        m = self.exterior_manifold()
        v = m.volume()
        return (float(v), v.accuracy, m.solution_type())

    def exterior_manifold(self):
        '''
        The knot complement manifold of self as a SnapPy class
        giving access to all of SnapPy's tools.

        This method requires that Spherogram, and possibly SnapPy,
        are installed.
        '''
        link = self.planar_diagram().as_spherogram()
        link.simplify()  # necessary with snappy 2.5, which can't deal
                         # with extra Reidemeister moves sometimes
        return link.exterior()

    def __str__(self):
        if self._crossings is not None:
            return '<Knot with {} points, {} crossings>'.format(
                len(self.points), len(self._crossings))
        return '<Knot with {} points>'.format(len(self.points))

    def identify(self, determinant=True, alexander=False, roots=(2, 3, 4),
                 min_crossings=True):
        '''
        Provides a simple interface to
        :func:`pyknotid.catalogue.identify.from_invariants`, by passing
        the given invariants. This does *not* support all invariants
        available, or more sophisticated identification methods,
        so don't be afraid to use the catalogue functions directly.

        Parameters
        ----------
        determinant : bool
            If True, uses the knot determinant in the identification.
            Defaults to True.
        alexander : bool
            If True-like, uses the full alexander polynomial in the
            identification. If the input is a dictionary of kwargs,
            these are passed straight to self.alexander_polynomial.
        roots : iterable
            A list of roots of unity at which to evaluate. Defaults
            to (2, 3, 4), the first of which is redundant with the
            determinant. Note that higher roots can be calculated, but
            aren't available in the database.
        min_crossings : bool
            If True, the output is restricted to knots with fewer crossings
            than the current projection of this one. Defaults to True. The
            only reason to turn this off is to see what other knots have
            the same invariants, it is never not useful for direct
            identification.
        '''
        if not roots:
            roots = []
        roots = set(roots)
        if determinant:
            roots.add(2)

        identify_kwargs = {}
        for root in roots:
            identify_kwargs[
                'alex_imag_{}'.format(root)] = self.alexander_at_root(root)

        if alexander:
            if not isinstance(alexander, dict):
                import sympy as sym
                alexander = {'variable': sym.var('t')}
            poly = self.alexander_polynomial(**alexander)
            identify_kwargs['alexander'] = poly

        if min_crossings and len(self.gauss_code()) < 16:
            identify_kwargs['max_crossings'] = len(self.gauss_code())

        from pyknotid.catalogue.identify import from_invariants
        return from_invariants(**identify_kwargs)

    def planar_writhe_quantities(self, num_angles=100, **kwargs):
        '''Returns the second order writhes, and arnold 2St+J+ values, for a
        range of different projection directions.

        '''
        from pyknotid.spacecurves.rotation import (
            get_rotation_angles, rotate_to_top)

        angles = get_rotation_angles(num_angles)

        results = []

        print_dist = int(max(1, 3000. / len(self.points)))
        # import matplotlib.pyplot as plt
        # fig, ax = plt.subplots()
        for i, angs in enumerate(angles):
            if i % print_dist == 0:
                self._vprint('\ri = {} / {}'.format(i, len(angles)), False)
            
            k = Knot(self.points, verbose=False)
            k._apply_matrix(rotate_to_top(*angs))

            results.append((angs, k, k.planar_second_order_writhe(),
                            k.arnold_2St_2Jplus()))

            # ax.clear()
            # k.plot_projection(fig_ax=(fig, ax))
            # fig.set_size_inches((3, 3))
            # ax.set_title('Wr2 = {}, unknot Wr2 = {}'.format(results[-1][2],
            #                                                 results[-1][3]))
            # fig.tight_layout()
            # fig.savefig('planar_writhe_quantities-{:05d}.png'.format(i))

        return results


    def slipknot_alexander(self, num_samples=0, **kwargs):
        '''
        Parameters
        ----------
        num_samples : int
            The number of indices to cut at. Defaults to 0, which
            means to sample at all indices.
        **kwargs :
            Keyword arguments, passed directly to
            :meth:`pyknotid.spacecurves.openknot.OpenKnot.alexander_fractions.
        '''
        points = self.points
        if num_samples == 0:
            num_samples = len(points)
        indices = n.linspace(0, len(points), num_samples).astype(n.int)

        from pyknotid.spacecurves.openknot import OpenKnot

        arr = n.ones((num_samples, num_samples))
        for index, points_index in enumerate(indices):
            self._vprint('\rindex = {} / {}'.format(index, len(indices)),
                         False)
            for other_index, other_points_index in enumerate(
                    indices[(index+2):]):
                k = OpenKnot(points[points_index:other_points_index],
                             verbose=False)
                if len(k.points) < 4:
                    alex = 1.
                else:
                    alex = k.alexander_fractions(**kwargs)[-1][0]
                arr[index, other_index] = alex

        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        ax.imshow(arr, interpolation='none')

        ax.plot(n.linspace(0, num_samples, 100) - 0.5,
                n.linspace(num_samples, 0, 100) - 0.5,
                color='black',
                linewidth=3)

        ax.set_xlim(-0.5, num_samples-0.5)
        ax.set_ylim(-0.4, num_samples-0.5)

        fig.show()

        return fig, ax

    def isolate_knot(self):
        '''Return indices of self.points within which the knot (if any)
        appears to lie, according to a simple closure algorithm.

        This method is experimental and may not provide very good results.
        '''
        if self._cached_isolated is not None:
            return self._cached_isolated
        determinant = self.determinant()

        from pyknotid.spacecurves import OpenKnot

        k1 = OpenKnot(self.points, verbose=False)
        start, end = _isolate_open_knot(k1, determinant, 0, len(k1))[1:]

        if end - start < 0.6 * len(k1):
            self._cached_isolated = (start, end)
            return start, end

        roll_dist = int(0.25*len(self.points))
        k2 = OpenKnot(n.roll(self.points, roll_dist, axis=0), verbose=False)
        start, end = _isolate_open_knot(k2, determinant, 0, len(k2))[1:]
        print('se', start, end)
        start -= roll_dist
        start %= len(self)
        end -= roll_dist
        end %= len(self)
        print('now', start, end)
        self._cached_isolated = (start, end)
        return start, end

    def plot_isolated(self, **kwargs):
        '''
        Plots the curve in red, except for the isolated local knot which
        is coloured blue. The local knot is found with self.isolate_knot,
        which may not be reliable or have good resolution.

        Parameters
        ==========
        **kwargs :
            kwargs are passed directly to :meth:`Knot.plot`.
        '''
        start, end = self.isolate_knot()
        if end < start:
            end, start = start, end
        mus = n.zeros(len(self.points))
        mus[start:end+1] = 0.4
        if end - start > 0.6*len(self) or end == start:
            mus = 0.4 - mus
        self.plot(mus=mus, **kwargs)

    def slip_triangle(self, func):

        from pyknotid.representations import Representation
        r = self.representation()

        length = len(r)

        cs = self.raw_crossings()

        results = {}

        invs = {}
        
        for i in range(length + 1):
            for j in range(length + 1):
                if i + j >= length:
                    continue

                new_r = Representation(r)

                points = self.points.copy()

                new_cs = cs.copy()

                new_start = 0
                new_end = -1
                for _ in range(i):
                    new_start = new_cs[0, 0] + 0.5*(new_cs[1, 0] - new_cs[0, 0])
                    new_cs = new_cs[new_cs[:, 1] != new_cs[0, 0]]
                    new_cs = new_cs[new_cs[:, 0] != new_cs[0, 0]]
                    new_r._remove_crossing(new_r._gauss_code[0][0, 0])
                for _ in range(j):
                    new_end = new_cs[-2, 0] + 0.5*(new_cs[-1, 0] - new_cs[-2, 0])
                    new_cs = new_cs[new_cs[:, 1] != new_cs[-1, 0]]
                    new_cs = new_cs[new_cs[:, 0] != new_cs[-1, 0]]
                    new_r._remove_crossing(new_r._gauss_code[0][-1, 0])

                new_points = points[int(n.ceil(new_start)):int(new_end)]
                new_start_remainder = new_start % 1
                new_end_remainder = new_end % 1

                new_points = n.vstack((points[int(new_start)] + new_start_remainder * (points[int(new_start) + 1] - points[int(new_start)]),
                                       new_points,
                                       points[int(new_end)] + new_end_remainder * (points[int(new_end) + 1] - points[int(new_end)])))
                # results[(i, j)] = points[int(new_start):int(new_end)]
                results[(i, j)] = new_points
                invs[(i, j)] = func(new_r)

        import matplotlib.pyplot as plt
        from matplotlib.gridspec import GridSpec

        points = self.points
        mins = n.min(self.points, axis=0)
        maxs = n.max(self.points, axis=0)
        span = max((maxs - mins)[:2])
        size = span * 1.16

        xmin = mins[0] - 0.08*span
        xmax = maxs[0] + 0.08*span
        ymin = mins[1] - 0.08*span
        ymax = maxs[1] + 0.08*span

        fig = plt.figure()
        grid = GridSpec(length, length)
        for coords, points in results.items():
            print('coords are', coords)
            ax = plt.subplot(grid[length - 1 - coords[0], coords[1]])
            ax.plot(points[:, 0], points[:, 1])
            ax.set_xticks([])
            ax.set_yticks([])
            ax.set_xlim(xmin, xmax)
            ax.set_ylim(ymin, ymax)

            inv = invs[coords]
            colour = 'red' if inv != 0 else 'green'
            ax.patch.set_facecolor(colour)
            ax.patch.set_alpha(0.1)

        fig.tight_layout()
        fig.show()
        return fig, ax

    def whitney_index(self):
        '''The degree of the Gauss map mapping a point on the curve to the
        direction of the positive tangent vector at this point.'''

        points = self.points
        tangents = self.tangents()

        directions = np.apply_along_axis(
            lambda arr: np.arctan2(arr[1], arr[0]), 1, tangents)

        directions[directions > np.pi] -= 2*np.pi
        directions[directions < -np.pi] += 2*np.pi

        jumps = np.roll(directions, -1) - directions


        index = np.sum(jumps > np.pi) - np.sum(jumps < -np.pi)

        return index

    def arnold_2St_2Jplus(self, **kwargs):
        from pyknotid.invariants import arnold_2St_2Jplus
        gc = self.gauss_code(**kwargs)
        # Do *not* simplify, as this is only a plane curve invariant
        return arnold_2St_2Jplus(gc)

    def arnold_2St_2Jminus(self, **kwargs):
        from pyknotid.invariants import arnold_2St_2Jminus
        gc = self.gauss_code(**kwargs)
        # Do *not* simplify, as this is only a plane curve invariant
        return arnold_2St_2Jminus(gc)

    def plot_secant_manifold(self, plot_extrema=True):
        import matplotlib.pyplot as plt
        # from colorsys import hls_to_rgb
        from hsluv import hsluv_to_rgb, hpluv_to_rgb
        from scipy.special import erfinv

        # import pyknotid.hsluvcolormap  # causes hsluv to be registered

        points = self.points

        colours = np.ones((len(points), len(points), 3))
        heights = np.zeros((len(points), len(points)))

        thetas = []
        phis = []

        for i1, point in enumerate(points):
            for i2, other_point in enumerate(points[i1+1:]):
                i2 += i1 + 1
                direction = other_point - point
                theta = np.arccos(direction[2] / mag(direction))
                phi = np.arctan2(direction[1], direction[0])

                # colours[i2, i1] = hls_to_rgb((phi + np.pi) / (2*np.pi), erfinv((theta / np.pi) * 2 - 1.) / 4.4 + 0.5, 1)
                colours[i2, i1] = hsluv_to_rgb((360 * (phi + np.pi) / (2*np.pi), 100, 65))# * (erfinv((theta / np.pi) * 2 - 1.) / 4.4 + 0.5)))

                colours[i1, i2] =  hsluv_to_rgb((360 * ((phi + 2*np.pi) % (2*np.pi)) / (2*np.pi), 100, 65))

                phis.append(phi)
                thetas.append(theta)

                heights[i2, i1] = theta
                heights[i1, i2] = np.pi - theta

        print(np.min(phis), np.max(phis), np.min(thetas), np.max(thetas))

        fig, ax = plt.subplots()

        im = ax.imshow(colours, origin='lower', zorder=-1)
        # ax.contour(heights, cmap='Greys', levels=np.linspace(-1, 1, 13))
        ax.contour(heights, cmap='Greys_r', levels=np.linspace(0, np.pi, 11),
                   zorder=0)

        crossings = self.raw_crossings()

        # Plot the lines between crossings
        unique_crossings = []
        crossings_done = set()
        for crossing in crossings:
            if crossing[0] in crossings_done:
                continue
            crossings_done.add(crossing[1])
            unique_crossings.append(crossing)

        for i, crossing in enumerate(unique_crossings):
            start1, end1, height1, sign1 = crossing
            for other_crossing in unique_crossings[i+1:]:
                start2, end2, height2, sign2 = other_crossing

                if not (start2 > start1 and end1 > start2 and end2 > end1):
                    continue

                colour = 'crimson' if sign1 * sign2 > 0 else 'lime'
                ax.plot([start1, start2], [end1, end2], color=colour, zorder=1)

        # Plot the crossing points
        for crossing in crossings:
            if crossing[1] > crossing[0]:
                colour = 'crimson' if crossing[3] > 0 else 'lime'
                edge_colour = 'white' if crossing[2] > 0 else 'black'
                ax.scatter([crossing[0]], [crossing[1]], color=colour, edgecolors=edge_colour,
                           s=30, zorder=2)

        # Plot extrema of the planar projection
        zs = points[:, 1]
        maxima_indices = np.argwhere((zs > np.roll(zs, 1)) & (zs > np.roll(zs, -1))).T[0]
        minima_indices = np.argwhere((zs < np.roll(zs, 1)) & (zs < np.roll(zs, -1))).T[0]
        print('maxima indices', maxima_indices)
        print('minima indices', minima_indices)
        for maximum in maxima_indices:
            ax.scatter([maximum], [maximum],  color='purple', edgecolors='pink', zorder=5)
        for minimum in minima_indices:
            ax.scatter([minimum], [minimum],  color='cyan', edgecolors='pink', zorder=5)

        xs = np.arange(len(points))
        ax.plot(xs, xs, linewidth=8, color='black', zorder=4)

        ax.set_xticks([])
        ax.set_yticks([])


        fig.tight_layout()
        
        # Plot the knot projection inset
        from mpl_toolkits.axes_grid1.inset_locator import inset_axes
        inset_ax = inset_axes(ax, width="45%", height="45%", loc=4)
        # inset_ax = fig.add_axes([0.6, 0.1, 0.4, 0.4])
        self.plot_projection(fig_ax=(fig, inset_ax))
        # inset_ax.set_axis_off()
        inset_ax.set_xticks([])
        inset_ax.set_yticks([])
        inset_ax.patch.set_alpha(0.5)

        ax.set_xlim(0.5, len(points) - 0.5)
        ax.set_ylim(0.5, len(points) - 0.5)

        ax.set_ylabel('i2')
        ax.set_xlabel('i1')

        return fig, ax

    def plot_secant_crossings(self, radius=None, **kwargs):
        
        crossings = self.raw_crossings(**kwargs)

        if radius is None:
            radii = self.points - np.average(self.points, axis=0)
            radius = 2.5 * np.max(np.sqrt(np.sum(radii**2, axis=1)))

        unique_crossings = []
        crossings_done = set()
        for crossing in crossings:
            if crossing[0] in crossings_done:
                continue
            crossings_done.add(crossing[1])
            unique_crossings.append(crossing)

        points = self.points

        lines = []

        for i, crossing in enumerate(unique_crossings):
            start1, end1, height1, sign1 = crossing
            for other_crossing in unique_crossings[i+1:]:
                start2, end2, height2, sign2 = other_crossing

                if not (start2 > start1 and end1 > start2 and end2 > end1):
                    continue

                start_point1 = points[int(start1)]
                start_point1 += (start1 - int(start1)) * (
                    points[int(start1) + 1] - points[int(start1)])
                segment_point1s = points[int(start1) + 1:int(start2) + 1]
                end_point1 = points[int(start2)]
                end_point1 += (start2 - int(start2)) * (
                    points[int(start2) + 1] - points[int(start2)])

                segment1 = np.vstack(
                    [start_point1, segment_point1s, end_point1])

                #
                start_point2 = points[int(end1)]
                start_point2 += (end1 - int(end1)) * (
                    points[int(end1) + 1] - points[int(end1)])
                segment_point2s = points[int(end1) + 1:int(end2) + 1]
                end_point2 = points[int(end2)]
                end_point2 += (end2 - int(end2)) * (
                    points[int(end2) + 1] - points[int(end2)])

                segment2 = np.vstack(
                    [start_point2, segment_point2s, end_point2])

                directions = np.vstack([
                    start_point1 - segment2,
                    segment1 - end_point2])

                lines.append(directions)

        sphere_lines = []

        for i, line in enumerate(lines):
            radii = np.sqrt(np.sum(line**2, axis=1))
            thetas = np.arccos(line[:, 2] / radii)
            phis = np.arctan2(line[:, 1], line[:, 0])

            local_radius = radius - 0.02*i*radius*np.sin(2*np.sin(thetas))

            sphere_points = np.zeros(line.shape)
            sphere_points[:, 0] = radius * np.sin(thetas) * np.cos(phis)
            sphere_points[:, 1] = radius * np.sin(thetas) * np.sin(phis)
            sphere_points[:, 2] = radius * np.cos(thetas)

            # sphere_points += 2*(np.random.random(sphere_points.shape) - 0.5) * 0.005 * radius

            sphere_lines.append(sphere_points)

        self.plot(tube_radius=0.1, colour=(0.3, 0.3, 0.3, 1))

        # Plot the poles
        from vispy.geometry import create_sphere
        from vispy.scene import Mesh
        meshdata = create_sphere(radius=0.035 * radius)
        vertices = meshdata.get_vertices()
        faces = meshdata.get_faces()
        vertices[:, 2] += radius
        mesh1 = Mesh(vertices=vertices, faces=faces, vertex_colors=np.array([(0.3, 0.3, 0.3, 1) for v in vertices]))
        vertices = vertices.copy()
        vertices[:, 2] -= 2*radius
        mesh2 = Mesh(vertices=vertices, faces=faces, vertex_colors=np.array([(0.3, 0.3, 0.3, 1) for v in vertices]))

        import pyknotid.visualise as pvis
        pvis.vispy_canvas.view.add(mesh1)
        pvis.vispy_canvas.view.add(mesh2)

        from colorsys import hsv_to_rgb
        colours = [hsv_to_rgb(hue, 1, 0.8) for hue in np.linspace(0, 1, len(sphere_lines) + 1)][:-1]
        colours = np.array(colours)
        np.random.shuffle(colours)
        for line, colour in zip(sphere_lines, colours):
            from pyknotid.spacecurves.openknot import OpenKnot
            k = OpenKnot(line)
            k.plot(clf=False, tube_radius=0.15, colour=colour, zero_centroid=False)


def mag(v):
    return n.sqrt(v.dot(v))


def _isolate_open_knot(k, det, start, end):
    from pyknotid.spacecurves import OpenKnot
    if len(k.points) < 10:  
        return (False, None, None)

    alexanders = k.alexander_fractions()
    main_det, main_frac = alexanders[-1]


    if main_det != det:
        return (False, None, None)

    half_len = len(k.points) // 2
    k1 = OpenKnot(k.points[:half_len], verbose=False)
    k1_knotted, k1_start, k1_end = _isolate_open_knot(
        k1, det, start, start + half_len - 1)
    if k1_knotted:
        return (True, k1_start, k1_end)

    k2 = OpenKnot(k.points[half_len:], verbose=False)
    k2_knotted, k2_start, k2_end = _isolate_open_knot(
        k2, det, start + half_len, end)
    if k2_knotted:
        return (True, k2_start, k2_end)

    quarter_len = len(k.points) // 4
    k3 = OpenKnot(k.points[quarter_len:(quarter_len + half_len)], verbose=False)
    k3_knotted, k3_start, k3_end = _isolate_open_knot(
        k3, det, start + quarter_len, start + quarter_len + half_len)
    if k3_knotted:
        return (True, k3_start, k3_end)

    k4 = OpenKnot(k.points[quarter_len:], verbose=False)
    k4_knotted, k4_start, k4_end = _isolate_open_knot(
        k4, det, start + quarter_len, end)
    if k4_knotted:
        return (True, k4_start, k4_end)

    k5 = OpenKnot(k.points[:-quarter_len], verbose=False)
    k5_knotted, k5_start, k5_end = _isolate_open_knot(
        k5, det, start, end - quarter_len)
    if k5_knotted:
        return (True, k5_start, k5_end)

    return (True, start, end)
    





from __future__ import print_function
import sys

import numpy as np
cimport numpy as np

cimport cython

from libc.math cimport abs, pow, sqrt as csqrt, floor

cpdef cython_higher_order_writhe(double [:, :] points,
                        double [:, :] contributions,
                        long [:] order):

    cdef long i1, i2, i3, i4
    cdef long [:] indices = np.zeros(4, dtype=np.int)

    cdef double writhe = 0.0

    for i1 in range(len(points) - 3):
        print('\rcython i1', i1, len(points) - 4, end='')
        sys.stdout.flush()
        indices[0] = i1
        for i2 in range(i1 + 1, len(points) - 1):
            indices[1] = i2
            for i3 in range(i2 + 1, len(points) - 1):
                indices[2] = i3
                for i4 in range(i3 + 1, len(points) - 1):
                    indices[3] = i4

                    writhe += (contributions[indices[order[0]],
                                             indices[order[1]]] *
                               contributions[indices[order[2]],
                                             indices[order[3]]])
    print()

    return writhe 


cpdef cython_second_order_writhes(double [:, :] points,
                                  double [:, :] contributions):

    cdef long i1, i2, i3, i4
    cdef long [:] indices = np.zeros(4, dtype=np.int)

    cdef double writhe_1 = 0.0
    cdef double writhe_2 = 0.0
    cdef double writhe_3 = 0.0

    for i1 in range(len(points) - 3):
        if i1 % 5 == 0:
            print('\rcython i1', i1, len(points) - 4, end='')
        sys.stdout.flush()
        indices[0] = i1
        for i2 in range(i1 + 1, len(points) - 1):
            indices[1] = i2
            for i3 in range(i2 + 1, len(points) - 1):
                indices[2] = i3
                for i4 in range(i3 + 1, len(points) - 1):
                    indices[3] = i4

                    writhe_1 += contributions[i1, i2] * contributions[i3, i4]
                    writhe_2 += contributions[i1, i3] * contributions[i2, i4]
                    writhe_3 += contributions[i1, i4] * contributions[i2, i3]
    print()

    return (writhe_1 / (2*np.pi)**2,
            writhe_2 / (2*np.pi)**2,
            writhe_3 / (2*np.pi)**2)

cpdef cython_second_order_writhes_no_basepoint(double [:, :] points,
                                               double [:, :] contributions):

    cdef long i1, i2, i3, i4
    cdef long [:] indices = np.zeros(4, dtype=np.int)

    cdef double writhe_1 = 0.0
    cdef double writhe_2 = 0.0
    cdef double writhe_3 = 0.0

    for i1 in range(len(points) - 1):
        if i1 % 5 == 0:
            print('\rnbp cython i1', i1, len(points) - 4, end='')
        sys.stdout.flush()
        indices[0] = i1
        possible_i2s = list(range(i1 + 1, len(points) - 1)) + list(range(i1 ))
        for i2 in possible_i2s:
            indices[1] = i2
            if i2 > i1:
                possible_i3s = list(range(i2 + 1, len(points) - 1)) + list(range(i1 ))
            else:
                possible_i3s = list(range(i2 + 1, i1 ))
            for i3 in possible_i3s:
                indices[2] = i3
                if i3 > i1:
                    possible_i4s = list(range(i3 + 1, len(points) - 1)) + list(range(i1 ))
                else:
                    possible_i4s = list(range(i3 + 1, i1 ))
                for i4 in possible_i4s:
                    # print('i1, i2, i3, i4 = {}, {}, {}, {}'.format(i1, i2, i3, i4))
                    indices[3] = i4

                    writhe_1 += contributions[i1, i2] * contributions[i3, i4]
                    writhe_2 += contributions[i1, i3] * contributions[i2, i4]
                    writhe_3 += contributions[i1, i4] * contributions[i2, i3]
    print()

    return (writhe_1 / (2*np.pi)**2,
            writhe_2 / (2*np.pi)**2,
            writhe_3 / (2*np.pi)**2)


# cpdef writhing_matrix(double [:, :] points):
#     for i1 in range(len(points) - 3):
#         print('\ri = {} / {}'.format(i1, len(points) - 4), end='')
#         sys.stdout.flush()
#         p1 = points[i1]
#         for i2 in range(i1 + 2, len(points) - 1):
#             p2 = points[i2]

#             p4 = points[i2 + 1]

#             r12 = p2 - p1
#             r13 = p3 - p1
#             r14 = p4 - p1
#             r23 = p3 - p2
#             r24 = p4 - p2
#             r34 = p4 - p3

#             n1 = np.cross(r13, r14)
#             n1 /= np.sqrt(np.sum(n1**2))

#             n2 = np.cross(r14, r24)
#             n2 /= np.sqrt(np.sum(n2**2))

#             n3 = np.cross(r24, r23)
#             if np.any(np.abs(n3) > 0.):
#                 n3 /= np.sqrt(np.sum(n3**2))

#             n4 = np.cross(r23, r13)
#             if np.any(np.abs(n4) > 0.):
#                 n4 /= np.sqrt(np.sum(n4**2))

#             if np.any(np.isnan(n1)):
#                 print('!!! nan')
#                 print(i1, i2)
#                 print(p1, p2, p3, p4)
#                 print('nan', r23, r13, np.cross(r23, r13))

#             # When the vectors are nearly the same, floating point
#             # errors can sometimes make the output a tiny bit higher
#             # than 1
#             t1, t2, t3, t4 = np.clip([n1.dot(n2),
#                                       n2.dot(n3),
#                                       n3.dot(n4),
#                                       n4.dot(n1)],
#                                      -1, 1)

#             writhe_contribution = (np.arcsin(t1) +
#                                    np.arcsin(t2) +
#                                    np.arcsin(t3) +
#                                    np.arcsin(t4))

#             if np.isnan(writhe_contribution):
#                 print()
#                 print('nan!')
#                 print(i1, i2, n1, n2, n3, n4, writhe_contribution)
#                 print(n1.dot(n2) > 1, n2.dot(n3) > 1, n3.dot(n4) > 1, n4.dot(n1) > 1)
#                 print(n1.dot(n2), np.arcsin(n1.dot(n2)))
#                 print(n2.dot(n3), np.arcsin(n2.dot(n3)))
#                 print(n3.dot(n4), np.arcsin(n3.dot(n4)))
#                 print(n4.dot(n1), np.arcsin(n4.dot(n1)))

#             writhe_contribution *= np.sign(np.cross(r34, r12).dot(r13))
    
#             contributions[i1, i2] = writhe_contribution
#             contributions[i2, i1] = writhe_contribution

#     return contributions



from distutils.core import setup
from distutils.extension import Extension
from Cython.Distutils import build_ext

import numpy

ext_modules = [
        Extension("chelpers", ["chelpers.pyx"], libraries=["m"]),
        ]

setup(
  name = 'chelpers',
  cmdclass = {'build_ext': build_ext},
  include_dirs = [numpy.get_include()],
  ext_modules = ext_modules
)




from matplotlib.colors import Colormap
from hsluv import hsluv_to_rgb

class HsluvColormap(Colormap):
    def __getitem__(self, index):
        return hsluv_to_rgb(360*index, 100, 65)

from matplotlib.cm import register_cmap
register_cmap(name='hsluv', cmap=HsluvColormap('hsluv'))



'''Named knots and links
=====================

Functions for making certain knots. The knots available are an
arbitrary selection based on known analytic forms.

Each of these functions returns a
:class:`~pyknotid.spacecurves.knot.Knot` or other appropriate pyknotid
space curve class.

API documentation
-----------------

'''

import numpy as n
import numpy as np

from pyknotid.spacecurves.knot import Knot

def unknot(num_points=100):
    '''Returns a simple circle.'''
    data = n.zeros((num_points, 3), dtype=n.float64)
    ts = n.linspace(0, 2*n.pi, num_points)
    data[:, 0] = 3*n.sin(ts)
    data[:, 1] = 3*n.cos(ts)
    data[:, 2] = n.sin(3*ts)
    return Knot(data)

def k3_1(num_points=100):
    '''Returns a particular trefoil knot conformation.'''
    data = n.zeros((num_points, 3), dtype=n.float64)
    ts = n.linspace(0, 2*n.pi, num_points)
    data[:, 0] = (2+n.cos(3*ts))*n.cos(2*ts)
    data[:, 1] = (2+n.cos(3*ts))*n.sin(2*ts)
    data[:, 2] = n.sin(3*ts)
    return Knot(data)
trefoil = k3_1


def k4_1(num_points=100):
    '''Returns a particular figure eight knot conformation.'''
    data = n.zeros((num_points, 3), dtype=n.float64)
    ts = n.linspace(0, 2*n.pi, num_points)
    data[:, 0] = (2+n.cos(2*ts))*n.cos(3*ts)
    data[:, 1] = (2+n.cos(2*ts))*n.sin(3*ts)
    data[:, 2] = n.sin(4*ts)
    return Knot(data)
figure_eight = k4_1

def lissajous(nx=3, ny=2, nz=7, px=0.7, py=0.2, pz=0., num_points=100):
    '''Returns a `Lissajous knot
    <https://en.wikipedia.org/wiki/Lissajous_knot>`__ with the given
    parameters.'''
    data = n.zeros((num_points, 3), dtype=n.float64)
    ts = n.linspace(0, 2*n.pi, num_points)
    data[:, 0] = n.cos(nx*ts+px)
    data[:, 1] = n.cos(ny*ts+py)
    data[:, 2] = n.cos(nz*ts+pz)
    return Knot(data)


def k5_2(num_points=100):
    '''Returns a Lissajous conformation of the knot 5_2.'''
    return lissajous(3, 2, 7, 0.7, 0.2, 0., num_points)
three_twist = k5_2

def k6_1(num_points=100):
    '''Returns a Lissajous conformation of the knot 6_1.'''
    return lissajous(3, 2, 5, 1.5, 0.2, 0., num_points)
stevedore = k6_1

def k3_1_composite_3_1(num_points=100):
    '''Returns a Lissajous conformation of the composite double trefoil 3_1
    # 3_1.
    '''
    return lissajous(3, 5, 7, 0.7, 1., 0., num_points)
square = k3_1_composite_3_1

def k8_21(num_points=100):
    '''Returns a Lissajous conformation of the knot 8_21.'''
    return lissajous(3, 4, 7, 0.1, 0.7, 0., num_points)
